#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# class stuff for circles
#

from __future__ import generators

import math

import entity
import tolerance
import point
import graphicobject
import style
import linetype
import color
import quadtree

class Circle(graphicobject.GraphicObject):
    """A base-class for Circles and Arcs

A Circle has two attributes:

center: A Point object
radius: The Circle's radius

A Circle has the following methods:

{get/set}Center(): Get/Set the center Point of a Circle.
{get/set}Radius(): Get/Set the radius of a Circle.
move(): Move the Circle.
circumference(): Get the Circle's circumference.
area(): Get the Circle's area.
mapCoords(): Find the nearest Point on the Circle to a coordinate pair.
inRegion(): Returns whether or not a Circle can be seen in a bounded area.
clone(): Return an indentical copy of a Circle.
    """

    __defstyle = style.Style(u'Default Circle Style',
                            linetype.Linetype(u'Solid', None),
                            color.Color(0xffffff),
                            1.0)

    messages = {
        'moved' : True,
        'center_changed' : True,
        'radius_changed' : True,
        }

    def __init__(self, center, radius, st=None, lt=None, col=None, th=None, **kw):
        """Initialize a Circle.

Circle(center, radius[, st, lt, col, th])

The center should be a Point, or a two-entry tuple of floats,
and the radius should be a float greater than 0.
        """
        _cp = center
        if not isinstance(_cp, point.Point):
            _cp = point.Point(center)
        _r = radius
        if not isinstance(_r, float):
            _r = float(radius)
        if not _r > 0.0:
            raise ValueError, "Invalid radius: %g" % _r
        _st = st
        if _st is None:
            _st = Circle.__defstyle
        graphicobject.GraphicObject.__init__(self, _st, lt, col, th, **kw)
        self.__radius = _r
        self.__center = _cp
        _cp.connect('moved', self, Circle.movePoint)
        _cp.storeUser(self)

    def __eq__(self, obj):
        """Compare a Circle to another for equality.
        """
        if not isinstance(obj, Circle):
            return False
        if obj is self:
            return True
        _val = False
        if self.__center == obj.getCenter():
            if abs(self.__radius - obj.getRadius()) < 1e-10:
                _val = True
        return _val

    def __ne__(self, obj):
        """Compare a Circle to another for inequality.
        """
        if not isinstance(obj, Circle):
            return True
        if obj is self:
            return False
        _val = True
        if self.__center == obj.getCenter():
            if abs(self.__radius - obj.getRadius()) < 1e-10:
                _val = False
        return _val

    def getValues(self):
        _pids = entity.Entity.getValues(self)
        _gv = self.getGraphicValues(Circle.__defstyle)
        return 'circle', _pids, _gv, self.__center.getID(), self.__radius

    def getCenter(self):
        """Return the center Point of the Circle.

getCenter()
        """
        return self.__center

    def setCenter(self, c):
        """Set the center Point of the Circle.

setCenter(c)

The argument must be a Point or a tuple containing
two float values.
        """
        if self.isLocked():
            raise RuntimeError, "Setting center not allowed - object locked."
        _cp = self.__center
        if not isinstance(c, point.Point):
            raise TypeError, "Invalid center point: " + `c`
        if _cp is not c:
            _cp.disconnect(self)
            _cp.freeUser(self)
            self.__center = c
            self.sendMessage('center_changed', _cp)
            c.connect('moved', self, Circle.movePoint)
            c.storeUser(self)
            if abs(_cp.x - c.x) > 1e-10 or abs(_cp.y - c.y) > 1e-10:
                self.sendMessage('moved', _cp.x, _cp.y, self.__radius)
            self.modified()

    center = property(getCenter, setCenter, None, "Circle center")

    def getRadius(self):
        """Return the radius of the the Circle.

getRadius()
        """
        return self.__radius

    def setRadius(self, radius):
        """Set the radius of the Circle.

setRadius(radius)

The argument must be float value greater than 0.
        """
        if self.isLocked():
            raise RuntimeError, "Setting radius not allowed - object locked."
        _r = radius
        if not isinstance(_r, float):
            _r = float(radius)
        if not _r > 0.0:
            raise ValueError, "Invalid radius: %g" % _r
        _cr = self.__radius
        if abs(_cr - _r) > 1e-10:
            self.__radius = _r
            self.sendMessage('radius_changed', _cr)
            _cx, _cy = self.__center.getCoords()
            self.sendMessage('moved', _cx, _cy, _cr)
            self.modified()

    radius = property(getRadius, setRadius, None, "Circle radius")

    def move(self, dx, dy):
        """Move a Circle.

move(dx, dy)

The first argument gives the x-coordinate displacement,
and the second gives the y-coordinate displacement. Both
values should be floats.
        """
        if self.isLocked():
            raise RuntimeError, "Setting radius not allowed - object locked."
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if abs(_dx) > 1e-10 or abs(_dy) > 1e-10:
            _x, _y = self.__center.getCoords()
            self.ignore('moved')
            self.__center.move(_dx, _dy)
            self.receive('moved')
            self.sendMessage('moved', _x, _y, self.__radius)
            self.modified()

    def circumference(self):
        """Return the circumference of the Circle.

circumference()
        """
        return 2.0 * math.pi * self.__radius

    def area(self):
        """Return the area enclosed by the Circle.

area()
        """
        return math.pi * pow(self.__radius, 2)

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the Circle to a coordinate pair.

mapCoords(x, y[, tol])

The function has two required arguments:

x: A Float value giving the x-coordinate
y: A Float value giving the y-coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by coordinate pair to
an actual Point on the Circle. If the distance between the actual
Point and the coordinates used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _cx, _cy = self.__center.getCoords()
        _r = self.__radius
        _dist = math.hypot((_x - _cx), (_y - _cy))
        if abs(_dist - _r) < _t:
            _angle = math.atan2((_y - _cy),(_x - _cx))
            _xoff = _r * math.cos(_angle)
            _yoff = _r * math.sin(_angle)
            return (_cx + _xoff), (_cy + _yoff)
        return None

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not an Circle exists within a region.

inRegion(xmin, ymin, xmax, ymax[, fully])

The first four arguments define the boundary. The optional
fifth argument "fully" indicates whether or not the Circle
must be completely contained within the region or just pass
through it.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        _xc, _yc = self.__center.getCoords()
        _r = self.__radius
        #
        # cheap test to see if circle cannot be in region
        #
        if (((_xc - _r) > _xmax) or
            ((_yc - _r) > _ymax) or
            ((_xc + _r) < _xmin) or
            ((_yc + _r) < _ymin)):
            return False
        _val = False
        _bits = 0
        #
        # calculate distances from center to region boundary
        #
        if abs(_xc - _xmin) < _r: _bits = _bits | 1 # left edge
        if abs(_xc - _xmax) < _r: _bits = _bits | 2 # right edge
        if abs(_yc - _ymin) < _r: _bits = _bits | 4 # bottom edge
        if abs(_yc - _ymax) < _r: _bits = _bits | 8 # top edge
        if _bits == 0:
            #
            # circle must be visible - the center is in
            # the region and is more than the radius from
            # each edge
            #
            _val = True
        else:
            #
            # calculate distance to corners of region
            #
            if math.hypot((_xc - _xmin), (_yc - _ymax)) < _r:
                _bits = _bits | 0x10 # upper left
            if math.hypot((_xc - _xmax), (_yc - _ymin)) < _r:
                _bits = _bits | 0x20 # lower right
            if math.hypot((_xc - _xmin), (_yc - _ymin)) < _r:
                _bits = _bits | 0x40 # lower left
            if math.hypot((_xc - _xmax), (_yc - _ymax)) < _r:
                _bits = _bits | 0x80 # upper right
            #
            # if all bits are set then distance from circle center
            # to region endpoints is less than radius - circle
            # entirely outside the region
            #
            _val = not ((_bits == 0xff) or fully)
        return _val

    def movePoint(self, p, *args):
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _cp = self.__center
        if p is not _cp:
            raise ValueError, "Point is not circle center: " + `p`
        _x, _y = _cp.getCoords()
        self.sendMessage('moved', _x, _y, self.__radius)
        self.modified()

    def clone(self):
        """Create an identical copy of a Circle

clone()
        """
        _cp = self.__center.clone()
        _st = self.getStyle()
        _lt = self.getLinetype()
        _col = self.getColor()
        _th = self.getThickness()
        return Circle(_cp, self.__radius, _st, _lt, _col, _th)

    def sendsMessage(self, m):
        if m in Circle.messages:
            return True
        return graphicobject.GraphicObject.sendsMessage(self, m)

    def clipToRegion(self, xmin, ymin, xmax, ymax):
        """Return the portions of a circle visible in a region.

clipToRegion(xmin, ymin, xmax, ymax)

This method returns a list of tuples. Each tuple contains two
float values representing arcs which are seen in the region. Each
tuple has the start angle and end angle.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _xc, _yc = self.__center.getCoords()
        _r = self.__radius
        _bits = 0
        _arcs = []
        #
        # calculate distances from center to region boundaries
        #
        if abs(_xc - _xmin) < _r: _bits = _bits | 1 # left edge
        if abs(_xc - _xmax) < _r: _bits = _bits | 2 # right edge
        if abs(_yc - _ymin) < _r: _bits = _bits | 4 # bottom edge
        if abs(_yc - _ymax) < _r: _bits = _bits | 8 # top edge
        #
        # test if the circle is entirely contained or entirely
        # outside the region
        #
        print "bits: %#02x" % _bits
        if _bits == 0:
            #
            # if the circle center is in region then the entire
            # circle is visible since the distance from the center
            # to any edge is greater than the radius. If the center
            # is not in the region then the circle is not visible in
            # the region because the distance to any edge is greater
            # than the radius, and so one of the bits should have been
            # set
            #
            if ((_xmin < _xc <_xmax) and (_ymin < _yc < _ymax)):
                print "circle completely inside region"
                _arcs.append((0.0, 360.0)) # fully in region
        else:
            #
            # calculate distance to corners of region
            #
            if math.hypot((_xc - _xmin), (_yc - _ymax)) < _r:
                _bits = _bits | 0x10 # upper left, NW corner
            if math.hypot((_xc - _xmax), (_yc - _ymin)) < _r:
                _bits = _bits | 0x20 # lower right, SE corner
            if math.hypot((_xc - _xmin), (_yc - _ymin)) < _r:
                _bits = _bits | 0x40 # lower left, SW corner
            if math.hypot((_xc - _xmax), (_yc - _ymax)) < _r:
                _bits = _bits | 0x80 # upper right, NE corner
            #
            # based on the bit pattern the various possible intersections
            # can be determined
            #
            # there is much room for optimization in here - many
            # of the distances from the center point to the region
            # edges and corners are calculated numerous times, the
            # square of these values are also repeatedly calculated ...
            #
            _rsqr = _r * _r
            # _rtd = 180.0/math.pi
            print "bits: %#02x" % _bits
            if _bits == 0x01: # circle crosses left edge twice
                print "circle crosses left edge twice"
                _yd = math.sqrt(_rsqr - pow((_xc - _xmin), 2))
                _yt = _yc + _yd
                _yb = _yc - _yd
                print "yt: %g; yb: %g" % (_yt, _yb)
                assert _yt < _ymax, "ytop > ymax"
                assert _yb > _ymin, "ybot < ymin"
                if (_ymin < _yc < _ymax): # must be true
                    _at = _calc_angle((_yt - _yc), (_xmin - _xc))
                    _ab = _calc_angle((_yb - _yc), (_xmin - _xc))
                    _arcs.append((_ab, ((360.0 - _ab) + _at)))
                    if _xc > _xmin: # circle inside region
                        print "circle center inside region"
                    else:
                        print "circle center outside"
                else:
                    if _yc < _ymin:
                        print "yc < ymin (%g < %g)" % (_yc, _ymin)
                    elif _yc > _ymax:
                        print "yc > ymax (%g > %g)" % (_yc, _ymax)
                    else:
                        print "unexpected y: (%g, %g, %g)" % (_ymin, _yc, _ymax)
            elif _bits == 0x02: # circle crosses right edge twice
                print "circle crosses right edge twice"
                _yd = math.sqrt(_rsqr - pow((_xc - _xmax), 2))
                _yt = _yc + _yd
                _yb = _yc - _yd
                print "yt: %g; yb: %g" % (_yt, _yb)
                assert _yt < _ymax, "ytop > ymax"
                assert _yb > _ymin, "ybot < ymin"
                if (_ymin < _yc < _ymax): # must be true
                    _at = _calc_angle((_yt - _yc), (_xmin - _xc))
                    _ab = _calc_angle((_yb - _yc), (_xmin - _xc))
                    _arcs.append((_at, (_ab - _at)))
                    if _xc < _xmax: # circle inside region
                        print "circle inside"
                    else:
                        print "circle outside"
                else:
                    if _yc < _ymin:
                        print "yc < ymin (%g < %g)" % (_yc, _ymin)
                    elif _yc > _ymax:
                        print "yc > ymax (%g > %g)" % (_yc, _ymax)
                    else:
                        print "unexpected y: (%g, %g, %g)" % (_ymin, _yc, _ymax)
            elif _bits == 0x04: # circle crosses bottom twice
                print "circle crosses bottom twice"
                _xd = math.sqrt(_rsqr - pow((_yc - _ymin), 2))
                _xr = _xc + _xd
                _xl = _xc - _xd
                print "xl: %g; xr: %g" % (_xl, _xr)
                assert _xr < _xmax, "xright > xmax"
                assert _xl > _xmin, "xeft < xmin"
                if (_xmin < _xc < _xmax): # must be true
                    _al = _calc_angle((_ymin - _yc), (_xl - _xc))
                    _ar = _calc_angle((_ymin - _yc), (_xr - _xc))
                    _arcs.append((_ar, (360.0 - _ar + _al)))
                    if _yc > _ymin: # circle inside region
                        print "circle inside"
                    else:
                        print "circle outside"
                else:
                    if _xc < _xmin:
                        print "xc < xmin (%g < %g)" % (_xc, _xmin)
                    elif _yc > _ymax:
                        print "xc > xmax (%g > %g)" % (_xc, _xmax)
                    else:
                        print "unexpected x: (%g, %g, %g)" % (_xmin, _xc, _xmax)
            elif _bits == 0x08: # circle crosses top twice
                print "circle crosses top twice"
                _xd = math.sqrt(_rsqr - pow((_yc - _ymax), 2))
                _xr = _xc + _xd
                _xl = _xc - _xd
                print "xl: %g; xr: %g" % (_xl, _xr)
                assert _xr < _xmax, "xright > xmax"
                assert _xl > _xmin, "xeft < xmin"
                if (_xmin < _xc < _xmax): # must be true
                    _al = _calc_angle((_ymax - _yc), (_xl - _xc))
                    _ar = _calc_angle((_ymax - _yc), (_xr - _xc))
                    _arcs.append((_al, (360.0 - _al + _ar)))
                    if _yc < _ymax: # circle inside region
                        print "circle inside"
                    else:
                        print "circle outside"
                else:
                    if _xc < _xmin:
                        print "xc < xmin (%g < %g)" % (_xc, _xmin)
                    elif _yc > _ymax:
                        print "xc > xmax (%g > %g)" % (_xc, _xmax)
                    else:
                        print "unexpected x: (%g, %g, %g)" % (_xmin, _xc, _xmax)
            elif _bits == 0x09: # circle crosses left and top twice
                print "circle crosses left and top twice"
                _xd = math.sqrt(_rsqr - pow((_yc - _ymax), 2))
                _xr = _xc + _xd
                _xl = _xc - _xd
                _yd = math.sqrt(_rsqr - pow((_xc - _xmin), 2))
                _yt = _yc + _yd
                _yb = _yc - _yd
                # top -> left
                _a1 = _calc_angle((_ymax - _yc), (_xl - _xc))
                _a2 = _calc_angle((_yt - _yc), (_xmin - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # left -> top
                _a1 = _calc_angle((_yb - _yc), (_xmin - _xc))
                _a2 = _calc_angle((_ymax - _yc), (_xr - _xc))
                _arcs.append((_a1, (360.0 - _a1 + _a2)))
                if ((_xmin < _xc < _xmax) and
                    (_ymin < _yc < _ymax)):
                    print "circle inside"
                else:
                    print "unexpected center for region: (%g, %g)" % (_xc, _yc)
            elif _bits == 0x0a: # circle crosses top and right twice
                print "circle crosses top and right twice"
                _xd = math.sqrt(_rsqr - pow((_yc - _ymax), 2))
                _xr = _xc + _xd
                _xl = _xc - _xd
                _yd = math.sqrt(_rsqr - pow((_xc - _xmax), 2))
                _yt = _yc + _yd
                _yb = _yc - _yd
                # top -> right
                _a1 = _calc_angle((_ymax - _yc), (_xl - _xc))
                _a2 = _calc_angle((_yb - _yc), (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # right -> top
                _a1 = _calc_angle((_yt - _yc), (_xmax - _xc))
                _a2 = _calc_angle((_ymax - _yc), (_xr - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if ((_xmin < _xc < _xmax) and
                    (_ymin < _yc < _ymax)):
                    print "circle inside"
                else:
                    print "unexpected center for region: (%g, %g)" % (_xc, _yc)
            elif _bits == 0x06: # circle crosses right and bottom twice
                print "circle crosses right and bottom twice"
                _xd = math.sqrt(_rsqr - pow((_yc - _ymin), 2))
                _xr = _xc + _xd
                _xl = _xc - _xd
                _yd = math.sqrt(_rsqr - pow((_xc - _xmax), 2))
                _yt = _yc + _yd
                _yb = _yc - _yd
                # right -> bottom
                _a1 = _calc_angle((_yt - _yc), (_xmax - _xc))
                _a2 = _calc_angle((_ymin - _yc), (_xl - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # bottom -> right
                _a1 = _calc_angle((_ymin - _yc), (_xr - _xc))
                _a2 = _calc_angle((_yb - _yc), (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if ((_xmin < _xc < _xmax) and
                    (_ymin < _yc < _ymax)):
                    print "circle inside"
                else:
                    print "unexpected center for region: (%g, %g)" % (_xc, _yc)
            elif _bits == 0x05: # circle crosses bottom and left twice
                print "circle crosses bottom and left twice"
                _xd = math.sqrt(_rsqr - pow((_yc - _ymin), 2))
                _xr = _xc + _xd
                _xl = _xc - _xd
                _yd = math.sqrt(_rsqr - pow((_xc - _xmin), 2))
                _yt = _yc + _yd
                _yb = _yc - _yd
                # left -> bottom
                _a1 = _calc_angle((_yb - _yc), (_xmin - _xc))
                _a2 = _calc_angle((_ymin - _yc), (_xl - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # bottom -> left
                _a1 = _calc_angle((_ymin - _yc), (_xr - _xc))
                _a2 = _calc_angle((_yt - _xc), (_xmin - _xc))
                _arcs.append((_a1, (360.0 - _a1 + _a2)))
                if ((_xmin < _xc < _xmax) and
                    (_ymin < _yc < _ymax)):
                    print "circle inside"
                else:
                    print "unexpected center for region: (%g, %g)" % (_xc, _yc)
            elif _bits == 0x0c: # circle crosses top and bottom twice
                print "circle crosses top and bottom twice"
                _xd = math.sqrt(_rsqr - pow((_yc - _ymax), 2))
                _xtr = _xc + _xd
                _xtl = _xc - _xd
                _xd = math.sqrt(_rsqr - pow((_yc - _ymin), 2))
                _xbr = _xc + _xd
                _xbl = _xc - _xd
                # top -> bottom
                _a1 = _calc_angle((_ymax - _yc), (_xtl - _xc))
                _a2 = _calc_angle((_ymin - _yc), (_xbl - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # bottom -> top
                _a1 = _calc_angle((_ymin - _yc), (_xbr - _xc))
                _a2 = _calc_angle((_ymax - _yc), (_xtr - _xc))
                _arcs.append((_a1, (360.0 - _a1 + _a2)))
                if (_ymin < _yc < _ymax): # needed?
                    print "circle inside region"
                elif _yc < _ymin:
                    print "circle below region"
                else:
                    print "circle above region"
            elif _bits == 0x03: # circle crosses left and right twice
                print "circle crosses left and right twice"
                _yd = math.sqrt(_rsqr - pow((_xc - _xmin), 2))
                _ylt = _yc + _yd
                _ylb = _yc - _yd
                _yd = math.sqrt(_rsqr - pow((_xc - _xmax), 2))
                _yrt = _yc + _yd
                _yrb = _yc - _yd
                # left -> right
                _a1 = _calc_angle((_ylb - _yc), (_xmin - _xc))
                _a2 = _calc_angle((_yrb - _yc), (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # right -> left
                _a1 = _calc_angle((_yrt - _yc), (_xmax - _xc))
                _a2 = _calc_angle((_ylt - _yc), (_xmin - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if (_xmin < _xc < _xmax):
                    print "circle inside region"
                elif _xc < _xmin:
                    print "circle left of region"
                else:
                    print "circle right of region"
            elif _bits == 0x0b: # circle through left, top, right twice
                print "circle through left & top & right twice"
                _xd = math.sqrt(_rsqr - pow((_yc - _ymax), 2))
                _xtr = _xc + _xd
                _xtl = _xc - _xd
                _yd = math.sqrt(_rsqr - pow((_xc - _xmin), 2))
                _ylt = _yc + _yd
                _ylb = _yc - _yd
                _yd = math.sqrt(_rsqr - pow((_xc - _xmax), 2))
                _yrt = _yc + _yd
                _yrb = _yc - _yd
                # top -> left
                _a1 = _calc_angle((_ymax - _yc), (_xtl - _xc))
                _a2 = _calc_angle((_ylt - _yc), (_xmin - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # left -> right
                _a1 = _calc_angle((_ylb - _yc), (_xmin - _xc))
                _a2 = _calc_angle((_yrb - _yc), (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # right > top
                _a1 = _calc_angle((_yrt - _yc), (_xmax - _xc))
                _a2 = _calc_angle((_ymax - _yc), (_xtr - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if (_xmin < _xc < _xmax):
                    print "circle inside region"
                elif _xc < _xmin:
                    print "xc < xmin (%g, %g)" % (_xc, _xmin)
                else:
                    print "xc > xmax (%g, %g)" % (_xc, _xmax)
                if _yc < _ymin:
                    print "yc < ymin (%g, %g)" % (_yc, _ymin)
                else:
                    if _yc > _ymax:
                        print "yc > ymax (%g, %g)" % (_yc, _ymax)
            elif _bits == 0x0e: # circle through top, right, bottom twice
                print "circle through top & right & bottom twice"
                _xd = math.sqrt(_rsqr - pow((_yc - _ymax), 2))
                _xtr = _xc + _xd
                _xtl = _xc - _xd
                _xd = math.sqrt(_rsqr - pow((_yc - _ymin), 2))
                _xbr = _xc + _xd
                _xbl = _xc - _xd
                _yd = math.sqrt(_rsqr - pow((_xc - _xmax), 2))
                _yrt = _yc + _yd
                _yrb = _yc - _yd
                # top -> bottom
                _a1 = _calc_angle((_ymax - _yc), (_xtl - _xc))
                _a2 = _calc_angle((_ymin - _yc), (_xbl - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # bottom -> right
                _a1 = _calc_angle((_ymin - _yc), (_xbr - _xc))
                _a2 = _calc_angle((_yrb - _yc), (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # right -> top
                _a1 = _calc_angle((_yrt - _yc), (_xmax - _xc))
                _a2 = _calc_angle((_ymax - _yc), (_xtr - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if (_ymin < _yc < _ymax):
                    print "circle inside region"
                elif _yc < _ymin:
                    print "yc < ymin (%g, %g)" % (_yc, _ymin)
                else:
                    print "yc > ymax (%g, %g)" % (_yc, _ymax)
                if _xc < _xmin:
                    print "xc < xmin (%g, %g)" % (_xc, _xmin)
                else:
                    if _xc > _xmax:
                        print "xc > xmax (%g, %g)" % (_xc, _xmax)
            elif _bits == 0x07: # circle though right, bottom, left twice
                print "circle through right & bottom & left twice"
                _xd = math.sqrt(_rsqr - pow((_yc - _ymin), 2))
                _xbr = _xc + _xd
                _xbl = _xc - _xd
                _yd = math.sqrt(_rsqr - pow((_xc - _xmax), 2))
                _yrt = _yc + _yd
                _yrb = _yc - _yd
                _yd = math.sqrt(_rsqr - pow((_xc - _xmin), 2))
                _ylt = _yc + _yd
                _ylb = _yc - _yd
                # right -> left
                _a1 = _calc_angle((_yrt - _yc), (_xmax - _xc))
                _a2 = _calc_angle((_ylt - _yc), (_xmin - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # left -> bottom
                _a1 = _calc_angle((_ylb - _yc), (_xmin - _xc))
                _a2 = _calc_angle((_ymin - _yc), (_xbl - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # bottom -> right
                _a1 = _calc_angle((_ymin - _yc), (_xbr - _xc))
                _a2 = _calc_angle((_yrb - _yc), (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if (_xmin < _xc < _xmax):
                    print "circle inside region"
                elif _xc < _xmin:
                    print "xc < xmin (%g, %g)" % (_xc, _xmin)
                else:
                    print "xc > xmax (%g, %g)" % (_xc, _xmax)
                if _yc < _ymin:
                    print "yc < ymin (%g, %g)" % (_yc, _ymin)
                else:
                    if _yc > _ymax:
                        print "yc > ymax (%g, %g)" % (_yc, _ymax)
            elif _bits == 0x0d: # circle through bottom, left, top twice
                print "circle through bottom & left & top twice"
                _xd = math.sqrt(_rsqr - pow((_yc - _ymin), 2))
                _xbr = _xc + _xd
                _xbl = _xc - _xd
                _xd = math.sqrt(_rsqr - pow((_yc - _ymax), 2))
                _xtr = _xc + _xd
                _xtl = _xc - _xd
                _yd = math.sqrt(_rsqr - pow((_xc - _xmin), 2))
                _ylt = _yc + _yd
                _ylb = _yc - _yd
                # bottom -> top
                _a1 = _calc_angle((_ymin - _yc), (_xbr - _xc))
                _a2 = _calc_angle((_ymax - _yc), (_xtr - _xc))
                _arcs.append((_a1, (360.0 - _a1 + _a2)))
                # top -> left
                _a1 = _calc_angle((_ymax - _yc), (_xtl - _xc))
                _a2 = _calc_angle((_ylt - _yc), (_xmin - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # left -> bottom
                _a1 = _calc_angle((_ylb - _yc), (_xmin - _xc))
                _a2 = _calc_angle((_ymin - _yc), (_xbl - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if (_ymin < _yc < _ymax):
                    print "circle inside region"
                elif _yc < _ymin:
                    print "yc < ymin (%g, %g)" % (_yc, _ymin)
                else:
                    print "yc > ymax (%g, %g)" % (_yc, _ymax)
                if _xc < _xmin:
                    print "xc < xmin (%g, %g)" % (_xc, _xmin)
                else:
                    if _xc > _xmax:
                        print "xc > xmax (%g, %g)" % (_xc, _xmax)
            elif _bits == 0x19: # circle through left, top, and NW corner
                print "circle through left and top with NW corner"
                _xd = math.sqrt(_rsqr - pow((_yc - _ymax), 2))
                _yd = math.sqrt(_rsqr - pow((_xc - _xmin), 2))
                _a1 = _calc_angle(-_yd, (_xmin - _xc))
                _a2 = _calc_angle((_ymax - _yc), _xd)
                if _xc > _xmax:
                    _arcs.append((_a1, (_a2 - _a1)))
                else:
                    _arcs.append((_a1, (360.0 - _a1 + _a2)))
                if ((_xmin < _xc < _xmax) and
                    (_ymin < _yc < _ymax)):
                    print "circle inside region"
                else:
                    print "circle outside region"
            elif _bits == 0x8a: # circle through right, top, and NE corner
                print "circle through right and top with NE corner"
                _xd = math.sqrt(_rsqr - pow((_yc - _ymax), 2))
                _yd = math.sqrt(_rsqr - pow((_xc - _xmax), 2))
                _a1 = _calc_angle((_ymax - _yc), -_xd)
                _a2 = _calc_angle(-_yd, (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if ((_xmin < _xc < _xmax) and
                    (_ymin < _yc < _ymax)):
                    print "circle inside region"
                else:
                    print "circle outside region"
            elif _bits == 0x26: # circle through right, bottom, and SE corner
                print "circle through right and bottom with SE corner"
                _xd = math.sqrt(_rsqr - pow((_yc - _ymin), 2))
                _yd = math.sqrt(_rsqr - pow((_xc - _xmax), 2))
                _a1 = _calc_angle(_yd, (_xmax - _xc))
                _a2 = _calc_angle((_ymin - _yc), -_xd)
                _arcs.append((_a1, (_a2 - _a1)))
                if ((_xmin < _xc < _xmax) and
                    (_ymin < _yc < _ymax)):
                    print "circle inside region"
                else:
                    print "circle outside region"
            elif _bits == 0x45: # circle through left, bottom, and SW corner
                print "circle through left and bottom with SW corner"
                _xd = math.sqrt(_rsqr - pow((_yc - _ymin), 2))
                _yd = math.sqrt(_rsqr - pow((_xc - _xmax), 2))
                _a1 = _calc_angle((_ymin - _yc), _xd)
                _a2 = _calc_angle(_yd, (_xmin - _xc))
                if _xc < _xmin:
                    _arcs.append((_a1, (_a2 - _a1)))
                else:
                    _arcs.append((_a1, (360.0 - _a1 + _a2)))
                if ((_xmin < _xc < _xmax) and
                    (_ymin < _yc < _ymax)):
                    print "circle inside region"
                else:
                    print "circle outside region"
            elif _bits == 0x9b: # circle through left, right, NE and NW corner
                print "circle through left and right with NE, NW corners"
                _yd = math.sqrt(_rsqr - pow((_xc - _xmin), 2))
                _a1 = _calc_angle(-_yd, (_xmin - _xc))
                _yd = math.sqrt(_rsqr - pow((_xc - _xmax), 2))
                _a2 = _calc_angle(-_yd, (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if _xc < _xmin:
                    print "x < xmin (%g < %g)" % (_xc, _xmin)
                elif _xc > _xmax:
                    print "x > xmax (%g > %g)" % (_xc, _xmax)
                else:
                    if _yc < _ymax:
                        print "circle center in region"
                    else:
                        print "circle center outside region"
            elif _bits == 0xae: # circle through top, botton, NE and SE corner
                print "circle through top and bottom with NE, SE corners"
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a1 = _calc_angle((_ymax - _yc), -_xd)
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a2 = _calc_angle((_ymin - _yc), -_xd)
                _arcs.append((_a1, (_a2 - _a1)))
                if _yc < _ymin:
                    print "y < ymin (%g < %g)" % (_yc, _ymin)
                elif _yc > _ymax:
                    print "y > ymax (%g > %g)" % (_yc, _ymax)
                else:
                    if _xc < _xmax:
                        print "circle center in region"
                    else:
                        print "circle center outside region"
            elif _bits == 0x67: # circle through left, right, SE and SW corner
                print "circle through left and right with SE, SW corners"
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a1 = _calc_angle(_yd, (_xmax - _xc))
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a2 = _calc_angle(_yd, (_xmin - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if _xc < _xmin:
                    print "x < xmin (%g < %g)" % (_xc, _xmin)
                elif _xc > _xmax:
                    print "x > xmax (%g > %g)" % (_xc, _xmax)
                else:
                    if _yc > _ymin:
                        print "circle center inside region"
                    else:
                        print "circle center outside region"
            elif _bits == 0x5d: # circle through top, bottom, SW and NW corner
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a1 = _calc_angle((_ymin - _yc), _xd)
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a2 = _calc_angle((_ymax - _yc), _xd)
                if _xc > _xmin:
                    _arcs.append((_a1, (_a2 - _a1)))
                else:
                    _arcs.append((_a1, (360.0 - _a1 + _a2)))
                print "circle through top and bottom with SW, NW corners"
                if _yc < _ymin:
                    print "y < ymin (%g < %g)" % (_yc, _ymin)
                elif _yc > _ymax:
                    print "y > ymax (%g > %g)" % (_yc, _ymax)
                else:
                    if _xc > _xmin:
                        print "circle center inside region"
                    else:
                        print "circle center outside region"
            elif _bits == 0xbf: # circle center NE, crosses left and bottom
                print "circle center NE of region, crosses left and bottom"
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a1 = _calc_angle(-_yd, (_xmin - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a2 = _calc_angle((_ymin - _yc), -_xd)
                _arcs.append((_a1, (_a2 - _a1)))
            elif _bits == 0xef: # circle center SE, crosses left and top
                print "circle center SE of region, crosses left and top"
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a1 = _calc_angle((_ymax - _yc), -_xd)
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a2 = _calc_angle(_yd, (_xmin - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
            elif _bits == 0x7f: # circle center SW, crosses top and right
                print "circle center SW of region, crosses top and right"
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a1 = _calc_angle(_yd, (_xmax - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a2 = _calc_angle((_ymax - _yc), _xd)
                _arcs.append((_a1, (_a2 - _a1)))
            elif _bits == 0xdf: # circle center NW, crosses right and bottom
                print "circle center NW of region, crosses right and bottom"
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a1 = _calc_angle((_ymin - _yc), _xd)
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a2 = _calc_angle(-_yd, (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
            elif _bits == 0x9f: # circle center N, crosses left, right, bottom
                print "circle center N, crosses left, right, and twice bottom"
                # left -> bottom
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a1 = _calc_angle(-_yd, (xmin - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a2 = _calc_angle((_ymin - _yc), -_xd)
                _arcs.append((_a1, (_a2 - _a1)))
                # bottom -> right; _xd now positive
                _a1 = _calc_angle((_ymin - _yc), _xd)
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a2 = _calc_angle(-_yd, (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if _yc < _ymax:
                    print "yc < ymax (%g < %g)" % (_yc, _ymax)
                if ((_xc <  _xmin) or (_xc > _xmax)):
                    print "xc: %g; xmin: %g; xmax: %g" % (_xc, _xmin, _xmax)
            elif _bits == 0xaf: # circle center W, crosses bottom, left, top
                print "circle center W, crosses bottom, top, and twice left"
                # top -> left
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a1 = _calc_angle((_ymax - _yc), -_xd)
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a2 = _calc_angle(_yd, (_xmin - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # left -> bottom; _yd now negative
                _a1 = _calc_angle(-_yd, (_xmin - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a2 = _calc_angle((_ymin - _yc), -_xd)
                _arcs.append((_a1, (_a2 - _a1)))
                if _xc < _xmax:
                    print "xc < xmax (%g < %g)" % (_xc, _xmax)
                if ((_yc < _ymin) or (_yc > _ymax)):
                    print "yc: %g; ymin: %g; ymax: %g" % (_yc, _ymin, _ymax)
            elif _bits == 0x6f: # circle center S, crosses left, top, right
                print "circle center S, crosses left, right, and twice top"
                # right -> top
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a1 = _calc_angle(_yd, (_xmax - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a2 = _calc_angle((_ymax - _yc), _xd)
                _arcs.append((_a1, (_a2 - _a1)))
                # top -> left; _xd now negative
                _a1 = _calc_angle((_ymax - _yc), -_xd)
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a2 = _calc_angle(_yd, (_xmin - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if _yc > _ymin:
                    print "yc > ymin (%g > %g)" % (_yc, _ymin)
                if ((_xc <  _xmin) or (_xc > _xmax)):
                    print "xc: %g; xmin: %g; xmax: %g" % (_xc, _xmin, _xmax)
            elif _bits == 0x5f: # circle center E, crosses top, right, bottom
                print "circle center E, crosses top, bottom, and twice right"
                # bottom -> right
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a1 = _calc_angle((_ymin - _yc), _xd)
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a2 = _calc_angle(-_yd, (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # right -> top, _yd now positive
                _a1 = _calc_angle(_yd, (_xmax - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a2 = _calc_angle((_ymax - _yc), _xd)
                _arcs.append((_a1, (_a2 - _a1)))
                if _xc > _xmin:
                    print "xc > xmin (%g > %g)" % (_xc, _xmin)
                if ((_yc < _ymin) or (_yc > _ymax)):
                    print "yc: %g; ymin: %g; ymax: %g" % (_yc, _ymin, _ymax)
            elif _bits == 0x1d:
                print "circle center N near NW, crosses T&L once, B twice"
                # left -> bottom
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a1 = _calc_angle(-_yd, (_xmin - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a2 = _calc_angle((_ymin - _yc), -_xd)
                _arcs.append((_a1, (_a2 - _a1)))
                # top -> bottom; _xd now positive
                _a1 = _calc_angle((_ymin - _yc), _xd)
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a2 = _calc_angle((_ymax - _yc), _xd)
                if _yc > _ymax:
                    _arcs.append((_a1, (_a2 - _a1)))
                else:
                    _arcs.append((_a1, (360.0 - _a1 + _a2)))
                if _xc < _xmin:
                    print "x < xmin (%g < %g)" % (_xc, _xmin)
                if _yc < _ymax:
                    print "y < ymax (%g < %g)" % (_yc, _ymax)
            elif _bits == 0x4d:
                print "circle center S near SW, crosses B&L once, T twice"
                # bottom -> top
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a1 = _calc_angle((_ymin - _yc), _xd)
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a2 = _calc_angle((_ymax - _yc), _xd)
                _arcs.append((_a1, (_a2 - _a1)))
                # top -> left, _xd now negative
                _a1 = _calc_angle((_ymax - _yc), -_xd)
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a2 = _calc_angle(_yd, (_xmin - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if _xc < _xmin:
                    print "x < xmin (%g < %g)" % (_xc, _xmin)
                if _yc > _ymin:
                    print "y > ymin (%g > %g)" % (_yc, _ymin)
            elif _bits == 0x2e:
                print "circle center S near SE, crosses B&R once, T twice"
                # right -> top
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a1 = _calc_angle(_yd, (_xmax - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a2 = _calc_angle((_ymax - _yc), _xd)
                _arcs.append((_a1, (_a2 - _a1)))
                # top -> bottom; _xd now negative
                _a1 = _calc_angle((_ymax - _yc), -_xd)
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a2 = _calc_angle((_ymin - _yc), -_xd)
                _arcs.append((_a1, (_a2 - _a1)))
                if _xc > _xmax:
                    print "x > xmax (%g > %g)" % (_xc, _xmax)
                if _yc > _ymin:
                    print "y > ymin (%g > %g)" % (_yc, _ymin)
            elif _bits == 0x8e:
                print "circle center N near NE, crosses T&R once, B twice"
                # top -> bottom
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a1 = _calc_angle((_ymax - _yc), -_xd)
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a2 = _calc_angle((_ymin - _yc), -_xd)
                _arcs.append((_a1, (_a2 - _a1)))
                # bottom -> right, _xd now positive
                _a1 = _calc_angle((_ymin - _yc), _xd)
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a2 = _calc_angle(-_yd, (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if _xc > _xmax:
                    print "x > xmax (%g > %g)" % (_xc, _xmax)
                if _yc < _ymax:
                    print "y < ymax (%g < %g)" % (_yc, _ymax)
            elif _bits == 0x1b:
                print "circle center E near NW, crosses T&L once, R twice"
                # left -> right
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a1 = _calc_angle(-_yd, (_xmin - _xc))
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a2 = _calc_angle(-_yd, (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # right -> top; _yd now positive
                _a1 = _calc_angle(_yd, (_xmax - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a2 = _calc_angle((_ymax - _yc), _xd)
                _arcs.append((_a1, (_a2 - _a1)))
                if _yc > _ymax:
                    print "y > ymax (%g > %g)" % (_yc, _ymax)
                if _xc > _xmin:
                    print "x > xmin (%g > %g)" % (_xc, _xmin)
            elif _bits == 0x8b:
                print "circle center W near NE, crosses T&R once, L twice"
                # top -> left
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a1 = _calc_angle((_ymax - _yc), -_xd)
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a2 = _calc_angle(_yd, (_xmin - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # left -> right; _yd now negative
                _a1 = _calc_angle(-_yd, (_xmin - _xc))
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a2 = _calc_angle(-_yd, (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if _yc > _ymax:
                    print "y > ymax (%g > %g)" % (_yc, _ymax)
                if _xc < _xmax:
                    print "x < xmax (%g < %g)" % (_xc, _xmax)
            elif _bits == 0x27:
                print "circle center W near SE, crosses B&R once, L twice"
                # right -> left
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a1 = _calc_angle(_yd, (_xmax - _xc))
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a2 = _calc_angle(_yd, (_xmin - _xc))
                print "a1: %g; a2: %g" % (_a1, _a2)
                _arcs.append((_a1, (_a2 - _a1)))
                # left -> bottom; now _yd is negative
                _a1 = _calc_angle(-_yd, (_xmin - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a2 = _calc_angle((_ymin - _yc), -_xd)
                print "a1: %g; a2: %g" % (_a1, _a2)
                _arcs.append((_a1, (_a2 - _a1)))
                if _yc < _ymin:
                    print "y < ymin (%g < %g)" % (_yc, _ymin)
                if _xc < _xmax:
                    print "x < xmax (%g < %g)" % (_xc, _xmax)
            elif _bits == 0x47:
                print "circle center E near SW, crosses B&L once, R twice"
                # bottom -> right
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a1 = _calc_angle((_ymin - _yc), _xd)
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a2 = _calc_angle(-_yd, (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # right -> left; now _yd is positive
                _a1 = _calc_angle(_yd, (_xmax - _xc))
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a2 = _calc_angle(_yd, (_xmin - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if _yc < ymin:
                    print "y < ymin (%g < %g)" % (_yc, _ymin)
                if _xc > _xmin:
                    print "x > xmin (%g > %g)" % (_xc, _xmin)
            elif _bits == 0x1f:
                print "circle center NW, crosses L&T once, R&B twice"
                # right -> top
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a1 = _calc_angle(_yd, (_xmax - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a2 = _calc_angle((_ymax - _yc), _xd)
                _arcs.append((_a1, (_a2 - _a1)))
                # left -> bottom
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a1 = _calc_angle(-_yd, (_xmin - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a2 = _calc_angle((_ymin - _yc), -_xd)
                _arcs.append((_a1, (_a2 - _a1)))
                # bottom -> right; _xd now positive
                _a1 = _calc_angle((_ymin - _yc), _xd)
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a2 = _calc_angle(-_yd, (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if ((_xmin < _xc < _xmax) and (_ymin < _yc < _ymax)):
                    print "circle center in region"
                else:
                    print "circle center out of region"
            elif _bits == 0x8f:
                print "circle center NE, crosses T&R once, B&L twice"
                # top -> left
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a1 = _calc_angle((_ymax - _yc), -_xd)
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a2 = _calc_angle(_yd, (_xmin - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # left -> bottom; _yd now negative
                _a1 = _calc_angle(-_yd, (_xmin - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a2 = _calc_angle((_ymin - _yc), -_xd)
                _arcs.append((_a1, (_a2 - _a1)))
                # bottom -> right; _xd now positive
                _a1 = _calc_angle((_ymin - _yc), _xd)
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a2 = _calc_angle(-_yd, (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if ((_xmin < _xc < _xmax) and (_ymin < _yc < _ymax)):
                    print "circle center in region"
                else:
                    print "circle center out of region"
            elif _bits == 0x2f:
                print "circle center SE, crosses L&T twice, R&B once"
                # top -> left
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a1 = _calc_angle((_ymax - _yc), -_xd)
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a2 = _calc_angle(_yd, (_xmin - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # left -> bottom; _yd now negative
                _a1 = _calc_angle(-_yd, (_xmin - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a2 = _calc_angle((_ymin - _yc), -_xd)
                _arcs.append((_a1, (_a2 - _a1)))
                # right -> top
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a1 = _calc_angle(_yd, (_xmax - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a2 = _calc_angle((_ymax - _yc), _xd)
                _arcs.append((_a1, (_a2 - _a1)))
                if ((_xmin < _xc < _xmax) and (_ymin < _yc < _ymax)):
                    print "circle center in region"
                else:
                    print "circle center out of region"
            elif _bits == 0x4f:
                print "circle center SW, crosses T&R twice, B&L once"
                # bottom -> right
                _xd = math.sqrt(_rsqr - pow((_ymin - _yc), 2))
                _a1 = _calc_angle((_ymin - _yc), _xd)
                _yd = math.sqrt(_rsqr - pow((_xmax - _xc), 2))
                _a2 = _calc_angle(-_yd, (_xmax - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # right -> top; _yd now positive
                _a1 = _calc_angle(_yd, (_xmax - _xc))
                _xd = math.sqrt(_rsqr - pow((_ymax - _yc), 2))
                _a2 = _calc_angle((_ymax - _yc), _xd)
                _arcs.append((_a1, (_a2 - _a1)))
                # top -> left; _xd now negative
                _a1 = _calc_angle((_ymax - _yc), -_xd)
                _yd = math.sqrt(_rsqr - pow((_xmin - _xc), 2))
                _a2 = _calc_angle(_yd, (_xmin - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                if ((_xmin < _xc < _xmax) and (_ymin < _yc < _ymax)):
                    print "circle center in region"
                else:
                    print "circle center out of region"
            elif _bits == 0x0f:
                print "circle crosses all edges twice"
                _yld = math.sqrt(_rsqr - pow((_xc - _xmin), 2))
                _yrd = math.sqrt(_rsqr - pow((_xc - _xmax), 2))
                _xtd = math.sqrt(_rsqr - pow((_yc - _ymax), 2))
                _xbd = math.sqrt(_rsqr - pow((_yc - _ymin), 2))
                # right -> top
                _x1 = _xmax
                _y1 = _yc + _yrd
                _x2 = _xc + _xtd
                _y2 = _ymax
                _a1 = _calc_angle((_y1 - _yc), (_x1 - _xc))
                _a2 = _calc_angle((_y2 - _yc), (_x2 - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # top -> left
                _x1 = _xc - _xtd
                _y1 = _ymax
                _x2 = _xmin
                _y2 = _yc + _yld
                _a1 = _calc_angle((_y1 - _yc), (_x1 - _xc))
                _a2 = _calc_angle((_y2 - _yc), (_x2 - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # left -> bottom
                _x1 = _xmin
                _y1 = _yc - _yld
                _x2 = _xc - _xbd
                _y2 = _ymin
                _a1 = _calc_angle((_y1 - _yc), (_x1 - _xc))
                _a2 = _calc_angle((_y2 - _yc), (_x2 - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
                # bottom -> right
                _x1 = _xc + _xbd
                _y1 = _ymin
                _x2 = _xmax
                _y2 = _yc - _yrd
                _a1 = _calc_angle((_y1 - _yc), (_x1 - _xc))
                _a2 = _calc_angle((_y2 - _yc), (_x2 - _xc))
                _arcs.append((_a1, (_a2 - _a1)))
            elif _bits == 0xff:
                print "circle outside region"
            else:
                print "Unexpected bit pattern: %#02x" % _bits
        return _arcs

def _calc_angle(dy, dx):
    _angle = math.atan2(dy, dx) * (180.0/math.pi)
    if _angle < 0.0:
        _angle = _angle + 360.0
    return _angle

#
# Quadtree Circle storage
#

class CircleQuadtree(quadtree.Quadtree):
    def __init__(self):
        quadtree.Quadtree.__init__(self)

    def getNodes(self, *args):
        _alen = len(args)
        if _alen != 3:
            raise ValueError, "Expected 3 arguments, got %d" % _alen
        _x, _y, _r = args        
        if not isinstance(_x, float):
            _x = float(args[0])
        if not isinstance(_y, float):
            _y = float(args[1])
        if not isinstance(_r, float):
            _r = float(args[2])
        _cxmin = _x - _r
        _cxmax = _x + _r
        _cymin = _y - _r
        _cymax = _y + _r
        _nodes = [self.getTreeRoot()]
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_cxmin > _xmax) or
                (_cxmax < _xmin) or
                (_cymin > _ymax) or
                (_cymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _cxmax < _xmid: # circle on left side
                    _ne = _se = False
                if _cxmin > _xmid: # circle on right side
                    _nw = _sw = False
                if _cymax < _ymid: # circle below
                    _nw = _ne = False
                if _cymin > _ymid: # circle above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                yield _node

    def addObject(self, obj):
        if not isinstance(obj, Circle):
            raise TypeError, "Invalid Circle object: " + `obj`
        if obj in self:
            return
        _x, _y = obj.getCenter().getCoords()
        _r = obj.getRadius()
        _bounds = self.getTreeRoot().getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _cxmin = _x - _r
        _cxmax = _x + _r
        _cymin = _y - _r
        _cymax = _y + _r
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _cxmin - 1.0
            _ymin = _cymin - 1.0
            _xmax = _cxmax + 1.0
            _ymax = _cymax + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _cxmin < _xmin:
                _xmin = _cxmin - 1.0
                _resize = True
            if _cxmax > _xmax:
                _xmax = _cxmax + 1.0
                _resize = True
            if _cymin < _ymin:
                _ymin = _cymin - 1.0
                _resize = True
            if _cymax > _ymax:
                _ymax = _cymax + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
        for _node in self.getNodes(_x, _y, _r):
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()            
            if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, CircleQuadtree.moveCircle)

    def old_addObject(self, obj):
        if not isinstance(obj, Circle):
            raise TypeError, "Invalid Circle object: " + `obj`
        if obj in self:
            return
        _x, _y = obj.getCenter().getCoords()
        _r = obj.getRadius()
        _node = self.getTreeRoot()
        _bounds = _node.getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _cxmin = _x - _r
        _cxmax = _x + _r
        _cymin = _y - _r
        _cymax = _y + _r
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _cxmin - 1.0
            _ymin = _cymin - 1.0
            _xmax = _cxmax + 1.0
            _ymax = _cymax + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _cxmin < _xmin:
                _xmin = _cxmin - 1.0
                _resize = True
            if _cxmax > _xmax:
                _xmax = _cxmax + 1.0
                _resize = True
            if _cymin < _ymin:
                _ymin = _cymin - 1.0
                _resize = True
            if _cymax > _ymax:
                _ymax = _cymax + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
            _node = self.getTreeRoot()
        _nodes = []
        _nodes.append(_node)
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_cxmin > _xmax) or
                (_cxmax < _xmin) or
                (_cymin > _ymax) or
                (_cymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _cxmax < _xmid: # circle on left side
                    _ne = _se = False
                if _cxmin > _xmid: # circle on right side
                    _nw = _sw = False
                if _cymax < _ymid: # circle below
                    _nw = _ne = False
                if _cymin > _ymid: # circle above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                    _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, CircleQuadtree.moveCircle)

    def delObject(self, obj):
        if obj not in self:
            return
        _x, _y = obj.getCenter().getCoords()
        _r = obj.getRadius()
        _pdict = {}
        for _node in self.getNodes(_x, _y, _r):
            _node.delObject(obj) # circle may not be in the node ...
            _parent = _node.getParent()
            if _parent is not None:
                _pid = id(_parent)
                if _pid not in _pdict:
                    _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def old_delObject(self, obj):
        if obj not in self:
            return
        _x, _y = obj.getCenter().getCoords()
        _r = obj.getRadius()
        _xmin = _ymin = _xmax = _ymax = None
        _cxmin = _x - _r
        _cxmax = _x + _r
        _cymin = _y - _r
        _cymax = _y + _r
        _pdict = {}
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_cxmin > _xmax) or
                (_cxmax < _xmin) or
                (_cymin > _ymax) or
                (_cymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _cxmax < _xmid: # circle on left side
                    _ne = _se = False
                if _cxmin > _xmid: # circle on right side
                    _nw = _sw = False
                if _cymax < _ymid: # circle below
                    _nw = _ne = False
                if _cymin > _ymid: # circle above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # circle may not be in the node ...
                _parent = _node.getParent()
                if _parent is not None:
                    _pid = id(_parent)
                    if _pid not in _pdict:
                        _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def find(self, *args):
        _alen = len(args)
        if _alen < 3:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _r = args[2]
        if not isinstance(_r, float):
            _r = float(args[2])
        _t = tolerance.TOL
        if _alen > 3:
            _t = tolerance.toltest(args[4])
        if not len(self):
            return None
        _cxmin = _x - _r
        _cxmax = _x + _r
        _cymin = _y - _r
        _cymax = _y + _r
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _cdict = {}
        _bailout = False
        _circ = None
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_cxmin > _xmax) or
                (_cxmax < _xmin) or
                (_cymin > _ymax) or
                (_cymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _cxmax < (_xmid - _t): # circle on left side
                    _ne = _se = False
                if _cxmin > (_xmid + _t): # circle on right side
                    _nw = _sw = False
                if _cymax < (_ymid - _t): # circle below
                    _nw = _ne = False
                if _cymin > (_ymid + _t): # circle above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                for _c in _node.getObjects():
                    _cid = id(_c)
                    if _cid not in _cdict:
                        _cx, _cy = _c.getCenter().getCoords()
                        if ((abs(_cx - _x) < _t) and (abs(_cy - _y) < _t)):
                            _cr = _c.getRadius()
                            if abs(_cr - _r) < _t:
                                _bailout = True
                                _circ = _c
                                break
                        _cdict[_cid] = True
            if _bailout:
                break
        return _circ

    def moveCircle(self, obj, *args):
        if obj not in self:
            raise ValueError, "Circle not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 3:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _r = args[2]
        if not isinstance(_r, float):
            _r = float(args[2])
        for _node in self.getNodes(_x, _y, _r):
            _node.delObject(obj) # circle may not be in node ...
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def old_moveCircle(self, obj, *args):
        if obj not in self:
            raise ValueError, "Circle not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 3:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _r = args[2]
        if not isinstance(_r, float):
            _r = float(args[2])
        _cxmin = _x - _r
        _cxmax = _x + _r
        _cymin = _y - _r
        _cymax = _y + _r
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_cxmin > _xmax) or
                (_cxmax < _xmin) or
                (_cymin > _ymax) or
                (_cymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _cxmax < _xmid: # circle on left side
                    _ne = _se = False
                if _cxmin > _xmid: # circle on right side
                    _nw = _sw = False
                if _cymax < _ymid: # circle below
                    _nw = _ne = False
                if _cymin > _ymid: # circle above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # circle may not be in node ...
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def getClosest(self, x, y, tol=tolerance.TOL):
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _circ = _tsep = None
        _bailout = False
        _cdict = {}
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_x < (_xmin - _t)) or
                (_x > (_xmax + _t)) or
                (_y < (_ymin - _t)) or
                (_y > (_ymax + _t))):
                continue
            if _node.hasSubnodes():
                _nodes.extend(_node.getSubnodes())
            else:
                for _c in _node.getObjects():
                    _cid = id(_c)
                    if _cid not in _cdict:
                        _cp = _c.mapCoords(_x, _y, _t)
                        if _cp is not None:
                            _cx, _cy = _cp
                            _sep = math.hypot((_cx - _x), (_cy - _y))
                            if _tsep is None:
                                _tsep = _sep
                                _circ = _c
                            else:
                                if _sep < _tsep:
                                    _tsep = _sep
                                    _circ = _c
                            if _sep < 1e-10 and _circ is not None:
                                _bailout = True
                                break
            if _bailout:
                break
        return _circ

    def getInRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _circs = []
        if not len(self):
            return _circs
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _cdict = {}
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                for _subnode in _node.getSubnodes():
                    _sxmin, _symin, _sxmax, _symax = _subnode.getBoundary()
                    if ((_sxmin > _xmax) or
                        (_symin > _ymax) or
                        (_sxmax < _xmin) or
                        (_symax < _ymin)):
                        continue
                    _nodes.append(_subnode)
            else:
                for _circ in _node.getObjects():
                    _cid = id(_circ)
                    if _cid not in _cdict:
                        if _circ.inRegion(_xmin, _ymin, _xmax, _ymax):
                            _circs.append(_circ)
                        _cdict[_cid] = True
        return _circs

#
# Circle history class
#

class CircleLog(graphicobject.GraphicObjectLog):
    def __init__(self, c):
        if not isinstance(c, Circle):
            raise TypeError, "Invalid circle: " + `c`
        graphicobject.GraphicObjectLog.__init__(self, c)
        c.connect('center_changed', self, CircleLog.centerChange)
        c.connect('radius_changed', self, CircleLog.radiusChange)

    def radiusChange(self, c, *args):
        _alen = len(args)
        if _alen < 1:
            raise ValueError, "Invalid argument count: %d" % _alen
        _r = args[0]
        if not isinstance(_r, float):
            _r = float(args[0])
        self.saveUndoData('radius_changed', _r)

    def centerChange(self, c, *args):
        _alen = len(args)
        if _alen < 1:
            raise ValueError, "Invalid argument count: %d" % _alen
        _old = args[0]
        if not isinstance(_old, point.Point):
            raise TypeError, "Invalid old center point: " + `_old`
        _oid = _old.getID()
        self.saveUndoData('center_changed', _oid)

    def execute(self, undo, *args):
        if undo is not True and undo is not False:
            raise ValueError, "Invalid undo value: " + str(undo)
        _alen = len(args)
        if _alen == 0:
            raise ValueError, "No arguments to execute()"
        _c = self.getObject()
        _cp = _c.getCenter()        
        _op = args[0]
        if _op == 'radius_changed':
            if len(args) < 2:
                raise ValueError, "Invalid argument count: %d" % _alen
            _r = args[1]
            if not isinstance(_r, float):
                _r = float(args[1])
            _cr = _c.getRadius()
            if undo:
                self.saveUndoData('radius_changed', _cr)
            else:
                self.saveRedoData('radius_changed', _cr)
            self.ignore('radius_changed')
            try:
                if undo:
                    _c.startUndo(True)
                    try:
                        _c.setRadius(_r)
                    finally:
                        _c.endUndo()
                    if abs(_r - _cr) > 1e-10:
                        _c.sendMessage('moved', _cp.x, _cp.y, _cr)
                else:
                    _c.setRadius(_r)
            finally:
                self.receive('radius_changed')
        elif _op == 'center_changed':
            if _alen < 2:
                raise ValueError, "Invalid argument count: %d" % _alen
            _oid = args[1]
            _cid = _cp.getID()
            if undo:
                self.saveRedoData('center_changed', _cid)
                self.ignore('center_changed')
                try:
                    _parent = _c.getParent()
                    if _parent is None:
                        raise ValueError, "Circle has no parent - cannot undo"
                    _op = _parent.getObject(_oid)
                    if _op is None or not isinstance(_op, point.Point):
                        raise ValueError, "Center point missing: id=%d" % _oid
                    _x1, _y1 = _cp.getCoords()
                    _c.startUndo(True)
                    try:
                        _c.setCenter(_op)
                    finally:
                        _c.endUndo()
                    if (abs(_op.x - _x1) > 1e-10 or
                        abs(_op.y - _y1) > 1e-10):
                        _c.sendMessage('moved', _x1, _y1, _c.getRadius())
                finally:
                    self.receive('center_changed')
            else:
                self.saveUndoData('center_changed', _cid)
                _parent = _c.getParent()
                if _parent is None:
                    raise ValueError, "Circle has no parent - cannot undo"
                _op = _parent.getObject(_oid)
                if _op is None or not isinstance(_op, point.Point):
                    raise ValueError, "Center point missing: id=%d" % _oid
                self.ignore('center_changed')
                try:
                    _c.setCenter(_op)
                finally:
                    self.receive('center_changed')
        else:
            raise ValueError, "Unexpected operation: %s" % _op



#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# classes for styles
#

import types

import color
import linetype

class Style(object):
    """A class storing a particular of Linetype, Color, and Thickness.

A Style consists of four attributes:

name: The Style name
linetype: A Linetype object
color: A Color object
thickness: A positive float value giving the line thickness

A Style has the following methods:

getName(): Get the Style name.
getColor(): Get the Style color.
getLinetype(): Get the Style Linetype.
getThickness(): Get the Style line thickness.
clone(): Return an identical copy of the Style.

Once a Style is created, the values in that object cannot
be changed.
    """
    def __init__(self, name, lt=None, col=None, t=None):
        """Instatiate a Style object.

Style(name [, lt, col, t])

name: A string giving the style a name

Option arguments:
lt: A Linetype object - defaults to a solid line Linetype
col: A Color object - defaults to the default Color object
t: A positive float value - defaults to 1.0
        """
        if not isinstance(name, types.StringTypes):
            raise TypeError, "Invalid Style name: " + `name`
        _n = name
        if not isinstance(_n, unicode):
            _n = unicode(name)
        _lt = lt
        if _lt is None:
            _lt = linetype.Linetype('Default_Solid', None)
        if not isinstance(_lt, linetype.Linetype):
            raise TypeError, "Invalid linetype: " + `_lt`
        _c = col
        if _c is None:
            _c = color.Color()
        if not isinstance(_c, color.Color):
            _c = color.Color(color)
        _t = t
        if _t is None:
            _t = 1.0
        if not isinstance(_t, float):
            _t = float(t)
        if _t < 0.0:
            raise ValueError, "Invalid line thickness: %g" % _t
        self.__name = _n
        self.__linetype = _lt
        self.__color = _c
        self.__thickness = _t

    def __eq__(self, obj):
        """Compare a Style object to another Style for equality.

Comparing two styles is really comparing that the linetypes
are the same, then the colors are the same, and lastly the
line thicness are the same (within a tiny tolerance). If all
three are the same, the comparison returns True. Otherwise, the
comparison returns False.
        """
        if not isinstance(obj, Style):
            raise TypeError, "Invalid object for Style comparison: " + str(obj)
        _val = False        
        if (self.__linetype == obj.getLinetype() and
            self.__color == obj.getColor() and
            abs(self.__thickness - obj.getThickness()) < 1e-10):
            _val = True
        return _val

    def __ne__(self, obj):
        """Compare a Style object to another Style for non-equality.

Comparing two styles is really comparing that the linetypes
are the same, then the colors are the same, and lastly the
line thicness are the same (within a tiny tolerance). If all
three are the same, the comparison returns False. Otherwise, the
comparison returns True.
        """
        if not isinstance(obj, Style):
            raise TypeError, "Invalid object for Style comparison: " + str(obj)
        _val = True        
        if (self.__linetype == obj.getLinetype() and
            self.__color == obj.getColor() and
            abs(self.__thickness - obj.getThickness()) < 1e-10):
            _val = False
        return _val

    def __hash__(self):
        """Return a hash value for the Style.

Defining this method allows Styles to be stored in dictionaries.
        """
        _val = hash(self.__color)
        _val = _val ^ hash(self.__linetype)
        _val = _val ^ hash(long(self.__thickness * 1e10))
        return _val

    def getName(self):
        """Return the name of the Style.

getName()        
        """
        return self.__name

    name = property(getName, None, None, "Style name.")
    
    def getLinetype(self):
        """Return the Linetype used by this Style.

getLinetype()        
        """
        return self.__linetype

    linetype = property(getLinetype, None, None, "Style Linetype")

    def getColor(self):
        """Return the Color used by this Style.

getColor()
        """
        return self.__color

    color = property(getColor, None, None, "Style Color")
    
    def getThickness(self):
        """Return the line thickness used by this style.

getThickness()        
        """
        return self.__thickness

    thickness = property(getThickness, None, None, "Style Thickness.")

    def getStyleValues(self):
        _n = self.__name
        _l = self.__linetype.getName(), self.__linetype.getList()
        _c = str(self.__color)
        _t = self.thickness
        return _n, _l, _c, _t
    
    def clone(self):
        """Return an identical copy of a Style.

clone()
        """
        _name = self.__name[:]
        _linetype = self.__linetype.clone()
        _color = self.__color.clone()
        _thickness = self.__thickness
        return Style(_name, _linetype, _color, _thickness)
#
# StyleDict Class
#
# The StyleDict is built from the dict object. Using instances
# of this class will guarantee than only Style objects will be
# stored in the instance
#

class StyleDict(dict):
    def __init__(self):
        dict.__init__(self)
        
    def __setitem__(self, key, value):
        if not isinstance(key, Style):
            raise TypeError, "StyleDict keys must be Style objects: " + `key`
        if not isinstance(value, Style):
            raise TypeError, "StyleDict values must be Style objects: " + `value`
        dict.__setitem__(self, key, value)



#
# Copyright (c) 2002-2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Handles creation of dimensions
#

import Generic.segment
import CocoaEntities

#
# Linear, Horizontal, Vertical dimensions all work the same way.
#
def ldim_mode_init(doc, tool):
    doc.setPrompt("Click on the first point for the dimension.")
    tool.setHandler("initialize", ldim_mode_init)
    tool.setHandler("left_button_press", ldim_first_left_button_press_cb)

def ldim_first_mouse_move_cb(doc, np, tool):
    _l1, _p1 = tool.getFirstPoint()
    _p2 = (np.x, np.y)
    _seg = Generic.segment.Segment(_p1, _p2)
    _da = doc.getDA()
    _da.setTempObject(_seg)

def ldim_second_mouse_move_cb(doc, np, tool):
    _ldim = tool.getDimension()
    _ldim.setLocation(np.x, np.y)
    _ldim.calcDimValues()
    _da = doc.getDA()
    _da.setTempObject(_ldim)

def ldim_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _layers = [_image.getTopLayer()]
    while len(_layers):
        _layer = _layers.pop()
        if _layer.isVisible():
            _pt = _layer.find('point', _viewLoc.x, _viewLoc.y, _tol)
            if _pt is not None:
                _x, _y = _pt.getCoords()
                _da.setTempPoint(_viewLoc)
                tool.setLocation(_x, _y)
                tool.setFirstPoint(_layer, _pt)
                tool.setHandler("left_button_press", ldim_second_left_button_press_cb)
                tool.setHandler("mouse_move", ldim_first_mouse_move_cb)
                doc.setPrompt("Click on the second point for the dimension.")
            break
        _layers.extend(_layer.getSublayers())

              
def ldim_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _layers = [_image.getTopLayer()]
    while len(_layers):
        _layer = _layers.pop()
        if _layer.isVisible():
            _pt = _layer.find('point', _viewLoc.x, _viewLoc.y, _tol)
            if _pt is not None:
                _x, _y = _pt.getCoords()
                tool.setSecondPoint(_layer, _pt)
                tool.setDimPosition(_x, _y)
                tool.clearCurrentPoint()
                tool.makeDimension(_image)
                tool.setHandler("left_button_press", ldim_third_left_button_press_cb)
                tool.setHandler("mouse_move", ldim_second_mouse_move_cb)
                doc.setPrompt("Click where the dimension text should be placed.")
                break
        _layers.extend(_layer.getSublayers())
            
def ldim_third_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    _ldim = tool.getDimension()
    _ldim.setLocation(_viewLoc.x, _viewLoc.y)
    _ldim.calcDimValues()
    _ldim.reset()
    CocoaEntities.create_entity(doc, tool)

#
# Radial
#
def radial_mode_init(doc, tool):
    doc.setPrompt("Click on an arc or a circle to dimension.")
    tool.setHandler("initialize", radial_mode_init)
    tool.setHandler("left_button_press", radial_first_left_button_press_cb)

def radial_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _layers = [_image.getTopLayer()]
    _dc = _dl = None
    while len(_layers):
        _layer = _layers.pop()
        if _layer.isVisible():
            _cobjs = (_layer.getLayerEntities("circle") +
                      _layer.getLayerEntities("arc"))
            for _cobj in _cobjs:
                _mp = _cobj.mapCoords(_viewLoc.x, _viewLoc.y, _tol)
                if _mp is not None:
                    _dc = _cobj
                    _dl = _layer
                    break
        _layers.extend(_layer.getSublayers())
    if _dc is not None:
        _x, _y = _mp
        tool.setDimObject(_dl, _dc)
        tool.setDimPosition(_x, _y)
        tool.makeDimension(_image)
        tool.setHandler("mouse_move", radial_mouse_move_cb)
        tool.setHandler("left_button_press", radial_second_left_button_press_cb)
        doc.setPrompt("Click where the dimension text should be placed.")

def radial_mouse_move_cb(doc, np, tool):
    _rdim = tool.getDimension()
    _rdim.setLocation(np.x, np.y)
    _rdim.calcDimValues()
    _da = doc.getDA()
    _da.setTempObject(_rdim)

def radial_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    _ldim = tool.getDimension()
    _ldim.setLocation(_viewLoc.x, _viewLoc.y)
    _ldim.calcDimValues()
    _ldim.reset()
    CocoaEntities.create_entity(doc, tool)
      

#
# Angular
#
def angular_mode_init(doc, tool):
    doc.setPrompt("Click on the angle vertex point or an arc.")
    tool.setHandler("initialize", angular_mode_init)
    tool.setHandler("left_button_press", angular_first_left_button_press_cb)

def angular_first_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    _viewLoc = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _layers = [_image.getTopLayer()]
    _pt = _arc = None
    while len(_layers):
        _layer = _layers.pop()
        if _layer.isVisible():
            _pt, _arc = _test_layer(_layer, _viewLoc.x, _viewLoc.y, _tol)
            if _pt is not None or _arc is not None:
                break
        _layers.extend(_layer.getSublayers())
    if _pt is not None:
        tool.setVertexPoint(_layer, _pt)
        _da.setTempPoint(_viewLoc)
        tool.setHandler("left_button_press", angular_second_left_button_press_cb)
        tool.setHandler("mouse_move", angular_segment_mouse_move_cb)
        doc.setPrompt("Click on the first endpoint for the dimension.")
    elif _arc is not None:
        _cp = _arc.getCenter()
        tool.setVertexPoint(_layer, _cp)
        _ep1, _ep2 = _arc.getEndpoints()
        _ex, _ey = _ep1
        _p1 = _layer.find('point', _ex, _ey)
        assert _p1 is not None, "Missing arc endpoint"
        tool.setFirstPoint(_layer, _p1)
        _ex, _ey = _ep2
        _p2 = _layer.find('point', _ex, _ey)
        assert _p2 is not None, "Missing arc endpoint"
        tool.setSecondPoint(_layer, _p2)
        tool.setDimPosition(_viewLoc.x, _viewLoc.y)
        tool.makeDimension(_image)
        tool.setHandler("left_button_press", angular_fourth_left_button_press_cb)
        tool.setHandler("mouse_move", angular_text_mouse_move_cb)
        doc.setPrompt("Click where the dimension text should be located.")
       
def angular_second_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _layers = [_image.getTopLayer()]
    while len(_layers):
        _layer = _layers.pop()
        if _layer.isVisible():
            _pt = _layer.find('point', _x, _y, _tol)
            if _pt is not None:
                _x, _y = _pt.getCoords()
                tool.setLocation(_x, _y)
                tool.setFirstPoint(_layer, _pt)
                tool.setHandler("left_button_press", angular_third_left_button_press_cb)
                tool.setHandler("mouse_move", angular_segment_mouse_move_cb)
                doc.setPrompt("Click on the second endpoint for the dimension.")
                break
        _layers.extend(_layer.getSublayers())

def angular_third_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    _tol = _da.pointSize().width
    _image = doc.getImage()
    _layers = [_image.getTopLayer()]
    while len(_layers):
        _layer = _layers.pop()
        if _layer.isVisible():
            _pt = _layer.find('point', _x, _y, _tol)
            if _pt is not None:
                _x, _y = _pt.getCoords()
                tool.setLocation(_x, _y)
                tool.setSecondPoint(_layer, _pt)
                tool.setDimPosition(_x, _y)
                tool.makeDimension(_image)
                tool.setHandler("left_button_press", angular_fourth_left_button_press_cb)
                tool.setHandler("mouse_move", angular_text_mouse_move_cb)
                doc.setPrompt("Click where the dimension text should be located.")
                break
        _layers.extend(_layer.getSublayers())
    
def angular_fourth_left_button_press_cb(doc, event, tool):
    _loc = event.locationInWindow()
    _da = doc.getDA()
    (_x, _y) = _da.convertPoint_fromView_(_loc, None)
    _adim = tool.getDimension()
    _adim.setLocation(_x, _y)
    _adim.calcDimValues()
    _adim.reset()
    CocoaEntities.create_entity(doc, tool)

def angular_segment_mouse_move_cb(doc, np, tool):
    _l1, _p1 = tool.getVertexPoint()
    _l2, _p2 = tool.getFirstPoint()
    _da = doc.getDA()
    _flag = True
    if _p2 is not None:
        _seg = Generic.segment.Segment(_p1, _p2)
        _da.setTempObject(_seg, _flag)
        _flag = False
    _p2 = (np.x, np.y)
    _seg = Generic.segment.Segment(_p1, _p2)
    _da.setTempObject(_seg, _flag)
    

def angular_text_mouse_move_cb(doc, np, tool):
    _adim = tool.getDimension()
    _adim.setLocation(np.x, np.y)
    _adim.calcDimValues()
    _da = doc.getDA()
    _da.setTempObject(_adim)
    
        
def _test_layer(layer, x, y, tol):
    _arc = None
    _pt = layer.find('point', x, y)
    if _pt is None:
        _pt = layer.find('point', x, y, tol)
        if _pt is None:
            _arc_pt = None
            for _arc in layer.getLayerEntities("arc"):
                _arc_pt = _arc.mapCoords(x, y, tol)
                if _arc_pt is not None:
                    break
            if _arc_pt is None:
                _arc = None # no hits on any arcs ...
    return _pt, _arc



#
# Copyright (c) 2002-2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Global strings/utility classes stored here
#

from Foundation import NSObject

# Strings for Preferences
WinFrameName = "CADWindow"
LayerColumnName = "LayerName"
LayerAddedNotification = "PCAddLayer"
LayerDeletedNotification = "PCDelLayer"
ToolChangedNotification ="PCToolChanged"

# Class to keep weak referenced/non-retained objects
# from causing crashes.
WRAPPED ={}

class ObjWrap(NSObject):
    """This class is here to wrap python objects as objective-c objects
    
    """
    def init_(self, obj):
        self.obj = obj
        return self
    
def wrap(item):
    if WRAPPED.has_key(item):
        return WRAPPED[item]
    else:
        WRAPPED[item] = ObjWrap.alloc().init_(item)
        return WRAPPED[item]

def unwrap(item):
    if item is None:
        return item 
    return item.obj
    
def purge(item):
    if WRAPPED.has_key(item):
        del WRAPPED[item]
    



#
# Copyright (c) 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# read autocad r12 file format
#

import struct
import array
import sys

r13 = False # ugh - global variable

def dump_info(handle):
    global r13
    handle.seek(0, 0)
    _version = handle.read(6)
    if _version != 'AC1009':
        if _version == 'AC1010':
            r13 = True
        else:
            print "unknown version"
            return
    handle.seek(6, 1) # skip rest of version
    (_b1, _w1, _w2, _w3, _b2) = struct.unpack("<BhhhB", handle.read(8))
    print "byte: %#02x" % _b2
    print "word: %d" % _w1
    print "word: %d" % _w2
    print "word: %d" % _w3
    print "byte: %#02x" % _b2
    _estart, _eend = struct.unpack("<ll", handle.read(8))
    print "entity offsets: %d to %d" % (_estart, _eend)
    _bsstart, _l1, _bsend, _l2 = struct.unpack("<llll", handle.read(16))
    print "blocksec offsets: %d to %d" % (_bsstart, _bsend)
    print "longs: %d, %d" % (_l1, _l2)
    _block_table = get_table(handle)
    print "block table: " + str(_block_table)
    _layer_table = get_table(handle)
    print "layer table: " + str(_layer_table)
    _style_table = get_table(handle)
    print "style table: " + str(_style_table)
    _ltype_table = get_table(handle)
    print "ltype table: " + str(_ltype_table)
    _view_table = get_table(handle)
    print "view table: " + str(_view_table)
    #
    read_header(handle)
    #
    # more tables ...
    _ucs_table = get_table(handle)
    print "ucs table: " + str(_ucs_table)
    handle.seek(0x500, 0)
    _vport_table = get_table(handle)
    print "vport table: " + str(_vport_table)
    handle.seek(8, 1)
    _appid_table = get_table(handle)
    print "appid table: " + str(_appid_table)
    handle.seek(6, 1)
    _dimstyle_table = get_table(handle)
    print "dimstyle table: " + str(_dimstyle_table)
    handle.seek(0x69f, 0)    
    _p13_table = get_table(handle)
    print "p13 table: " + str(_p13_table)
    handle.seek(38, 1)
    _at = handle.tell()
    print "handle offset: %d [%#x]" % (_at, _at)
    if _at != _estart:
        print "expected to be at entity start: %#x" % _estart
        return
    read_entities(handle, _estart, _eend)
    handle.seek(19, 1)
    read_block_table(handle, _block_table)
    read_layer_table(handle, _layer_table)
    read_style_table(handle, _style_table)
    read_ltype_table(handle, _ltype_table)
    read_view_table(handle, _view_table)
    read_ucs_table(handle, _view_table)
    read_vport_table(handle, _vport_table)
    read_appid_table(handle, _appid_table)
    read_dimstyle_table(handle, _dimstyle_table)
    read_p13_table(handle, _p13_table)
    read_entities(handle, _bsstart, _bsend)
    handle.seek(36, 1)
    _pestart, _peend, _pbstart,_pbend = struct.unpack('<4l', handle.read(16))
    print "pestart: %d; peend: %d" % (_pestart, _peend)
    print "bstart: %d; bend: %d" % (_pbstart, _pbend)    
    if (_pestart != _estart):
        print "pestart != estart: %d != %d" % (_pestart, _estart)
    if (_peend != _eend):
        print "peend != eend: %d != %d" % (_peend, _eend)
    if (_bsstart != _pbstart):
        print "pstart != bstart: %d != %d" % (_pbstart, _bsstart)
    if (_bsend != _pbend):
        print "pbend != bend: %d != %d" % (_pbend, _bsend)
    print "handle at: %#x" % handle.tell()        
    handle.seek(12, 1)
    print "handle at: %#x" % handle.tell()
    _bts = get_table_test(handle)
    print "test table: " + str(_bts)
    # handle.seek(6, 1)
    _lyrts = get_table_test(handle)
    print "test table: " + str(_lyrts)    
    # handle.seek(6, 1)
    _sts = get_table_test(handle)
    print "test table: " + str(_sts)    
    # handle.seek(6, 1)
    _ltts = get_table_test(handle)
    print "test table: " + str(_ltts)
    # handle.seek(6, 1)
    _vts = get_table_test(handle)
    print "test table: " + str(_vts)
    # handle.seek(6, 1)
    _uts = get_table_test(handle)
    print "test table: " + str(_uts)
    # handle.seek(6, 1)
    _vpts = get_table_test(handle)
    print "test table: " + str(_vts)
    # handle.seek(6, 1)
    _ats = get_table_test(handle)
    print "test table: " + str(_ats)
    # handle.seek(6, 1)
    _dts = get_table_test(handle)
    print "test table: " + str(_dts)
    # handle.seek(6, 1)
    _pts = get_table_test(handle)
    print "test table: " + str(_pts)
    # handle.seek(6, 1)
    
#
# entity reading routines
#

def line_reader(handle, flags, opts):
    print "line_reader() ..."
    _zflag = True
    if (flags & 0x4):
        _zflag = False
    _pt10 = get_point(handle, _zflag)
    _pt11 = get_point(handle, _zflag)
    print "p1: " + str(_pt10)
    print "p2: " + str(_pt11)
    if (opts & 0x1):
        _pt210 = get_point(handle, True)
        print "pt: " + str(_pt210)
    if (opts & 0x2):
        _db38 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db38
    _at = handle.tell()
    print "handle offset: %d [%#x]" % (_at, _at)

def point_reader(handle, flags, opts):
    print "point_reader() ..."
    _zflag = True
    if (flags & 0x4):
        _zflag = False
    _pt10 = get_point(handle, _zflag)
    print "point: " + str(_pt10)
    if (opts & 0x1):
        _pt210 = get_point(handle, True)
        print "pt: " + str(_pt210)
    if (opts & 0x2):
        _db38 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db38
    _at = handle.tell()
    print "handle offset: %d [%#x]" % (_at, _at)

def circle_reader(handle, flags, opts):
    print "circle_reader() ..."
    _zflag = False
    if (flags & 0x4):
        _zflag = True
    _pt10 = get_point(handle, _zflag)
    print "center: " + str(_pt10)
    _d40 = struct.unpack("<d", handle.read(8))[0]
    print "radius: %g" % _d40
    if (opts & 0x1):
        _pt210 = get_point(handle, True)
        print "pt: " + str(_pt210)
    if (opts & 0x2):
        _db38 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db38
    _at = handle.tell()
    print "handle offset: %d [%#x]" % (_at, _at)

def shape_reader(handle, flags, opts):
    print "shape_reader() ..."
    _pt10 = get_point(handle, False)
    print "pt: " + str(_pt10)
    _w2 = struct.unpack('<h', handle.read(2))[0]
    print "word: %d" % _w2
    if (opts & 0x1):
        _pt210 = get_point(handle, True)
        print "pt: " + str(_pt210)
    if (opts & 0x2):
        _db38 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db38
        
def text_reader(handle, flags, opts):
    print "text_reader() ..."
    _pt10 = get_point(handle, False)
    print "point: " + str(_pt10)
    _db40 = struct.unpack('<d', handle.read(8))[0]
    print "double: %g" % _db40 # angle of text
    _text = get_string(handle)
    print "text: %s" % _text
    if (opts & 0x1):
        _db50 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db50
    if (opts & 0x2):
        _db41 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db41
    if (opts & 0x4):
        _db51 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db51
    if (opts & 0x8):
        _b7 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b7
    if (opts & 0x10):
        _b71 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b71
    if (opts & 0x20):
        _b72 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b72
    if (opts & 0x40):
        _pt11 = get_point(handle, False)
        print "pt11: " + str(_pt11)
    if (opts & 0x100):
        _b73 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b73
    _at = handle.tell()
    print "handle offset: %d [%#x]" % (_at, _at)

def arc_reader(handle, flags, opts):
    print "arc_reader() ..."
    _pt10 = get_point(handle, False)
    print "center: " + str(_pt10)
    _d40 = struct.unpack('<d', handle.read(8))[0]
    print "radius: %g" % _d40
    _d50 = struct.unpack('<d', handle.read(8))[0]
    print "start angle: %g" % _d50
    _d51 = struct.unpack('<d', handle.read(8))[0]
    print "end angle: %g" % _d51
    if (opts & 0x1):
        _pt210 = get_point(handle, True)
        print "pt: " + str(_pt210)
    if (opts & 0x2):
        _db38 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db38

def trace_reader(handle, flags, opts):
    print "trace_reader() ..."
    _pt10 = get_point(handle, False)
    print "pt: " + str(_pt10)
    _pt11 = get_point(handle, False)
    print "pt: " + str(_pt11)
    _pt12 = get_point(handle, False)
    print "pt: " + str(_pt12)
    _pt13 = get_point(handle, False)
    print "pt: " + str(_pt13)
    if (opts & 0x1):
        _pt210 = get_point(handle, True)
        print "pt: " + str(_pt210)
    if (opts & 0x2):
        _db38 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db38

def solid_reader(handle, flags, opts):
    print "solid_reader() ..."
    _pt11 = get_point(handle, False)
    print "pt: " + str(_pt11)
    _pt12 = get_point(handle, False)
    print "pt: " + str(_pt12)
    _pt13 = get_point(handle, False)
    print "pt: " + str(_pt13)
    _pt14 = get_point(handle, False)
    print "pt: " + str(_pt14)
    if (opts & 0x1):
        _pt210 = get_point(handle, True)
        print "pt: " + str(_pt210)
    if (opts & 0x2):
        _db38 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db38

def block_reader(handle, flags, opts):
    print "block_reader() ..."
    _pt10 = get_point(handle, False)
    print "pt: " + str(_pt10)
    _s1 = get_string(handle)
    print "name: %s" % _s1
    if (opts & 0x2):
        _s3 = get_string(handle)
        print "string: %s" % _s3

def endblk_reader(handle, flags, opts):
    print "endblk_reader() ..."

def insert_reader(handle, flags, opts):
    print "insert_reader() ..."
    _w1 = struct.unpack('<h', handle.read(2))[0]
    print "word: %d" % _w1
    _p10 = get_point(handle, False)
    print "point: " + str(_p10)
    if (opts & 0x1):
        _db41 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db41
    if (opts & 0x2):
        _db42 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db42
    if (opts & 0x4):
        _db43 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db43
    if (opts & 0x8):
        _db50 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db50
    if (opts & 0x10):
        _w70 = struct.unpack('<h', handle.read(2))[0]
        print "word: %d" % _w70
    if (opts & 0x10):
        _w71 = struct.unpack('<h', handle.read(2))[0]
        print "word: %d" % _w71
    if (opts & 0x40):
        _db44 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db44
    if (opts & 0x80):
        _db45 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _db45
        
def attdef_reader(handle, flags, opts):
    print "attdef_reader() ..."
    _p10 = get_point(handle, False)
    print "pt: " + str(_p10)
    _s1 = get_string(handle)
    print "string: %s" % _s1
    _s3 = get_string(handle)
    print "string: %s" % _s3
    _s2 = get_string(handle)
    print "string: %s" % _s2
    _b70 = struct.unpack('B', handle.read(1))[0]
    print "byte: %#x" % _b70
    if (opts & 0x1):
        _b73 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b73
    if (opts & 0x2):
        _d50 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d50
    if (opts & 0x4):
        _d41 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d41
    if (opts & 0x8):
        _d42 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d42
    if (opts & 0x10):
        _b7 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b7
    if (opts & 0x20):
        _b71 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b71
    if (opts & 0x40):
        _b72 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b72
    if (opts & 0x80):
        _p11 = get_point(handle, False)
        print "pt: " + str(_p11)
    if (opts & 0x100):
        _p210 = get_point(handle, True)
        print "pt: " + str(_p210)
    if (opts & 0x200):
        _d38 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d38

def attrib_reader(handle, flags, opts):
    print "attrib_reader() ..."
    _p10 = get_point(handle, False)
    print "pt: " + str(_p10)
    _d40 = struct.unpack('<d', handle.read(8))[0]
    print "double: %g" % _d40
    _s1 = get_string(handle)
    print "string: %s" % _s1
    _s2 = get_string(handle)
    print "string: %s" % _s2
    if (opts & 0x1):
        _b73 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b73
    if (opts & 0x2):
        _d50 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d50
    if (opts & 0x4):
        _d41 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d41
    if (opts & 0x8):
        _d42 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d42
    if (opts & 0x10):
        _b7 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b7
    if (opts & 0x20):
        _b71 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b71
    if (opts & 0x40):
        _b72 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b72
    if (opts & 0x80):
        _p11 = get_point(handle, False)
        print "pt: " + str(_p11)
    if (opts & 0x100):
        _p210 = get_point(handle, True)
        print "pt: " + str(_p210)
    if (opts & 0x200):
        _d38 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d38

def sbend_reader(handle, flags, opts):
    print "sbend_reader() ..."
    _l = struct.unpack('<l', handle.read(4))[0]
    print "long: %d" % _l
    
def pline_reader(handle, flags, opts):
    print "pline_reader() ..."
    if (opts & 0x1):
        _b70 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b70
    if (opts & 0x2):
        _d40 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d40
    if (opts & 0x4):
        _b71 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b71
    if (opts & 0x8):
        _b72 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b72
    if (opts & 0x10):
        _b73 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b73
    if (opts & 0x20):
        _b74 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b74
    if (opts & 0x40):
        _b75 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b75

def vertex_reader(handle, flags, opts):
    print "vertex_reader() ..."
    _pt10 = get_point(handle, False)
    print "pt: " + str(_pt10)
    if (opts & 0x1):
        _d40 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d40
    if (opts & 0x2):
        _d41 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d41
    if (opts & 0x4):
        _d50 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d50
    if (opts & 0x8):
        _b70 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b70
    
def face3d_reader(handle, flags, opts):
    print "face3d_reader() ..."
    _zflag = False
    if (flags & 0x4):
        _zflag = True
    _pt10 = get_point(handle, _zflag)
    print "pt: " + str(_pt10)
    _pt11 = get_point(handle, _zflag)
    print "pt: " + str(_pt11)
    _pt12 = get_point(handle, _zflag)
    print "pt: " + str(_pt12)
    _pt13 = get_point(handle, _zflag)
    print "pt: " + str(_pt13)

def dim_reader(handle, flags, opts):
    print "dim_reader() ..."
    _w1 = struct.unpack('<h', handle.read(2))[0]
    print "word: %d" % _w1
    _pt10 = get_point(handle, True)
    print "pt: " + str(_pt10)
    _pt11 = get_point(handle, False)
    print "pt: " + str(_pt11)
    if (opts & 0x2):
        _b70 = struct.unpack('B', handle.read(1))[0]
        print "byte: %#x" % _b70
    if (opts & 0x1):
        _pt12 = get_point(handle, True)
        print "pt: " + str(_pt12)
    if (opts & 0x4):
        _s1 = get_string(handle)
        print "text: %s" % _s1
    if (opts & 0x8):
        _pt13 = get_point(handle, True)
        print "pt: " + str(_pt13)
    if (opts & 0x10):
        _pt14 = get_point(handle, True)
        print "pt: " + str(_pt14)
    if (opts & 0x20):
        _pt15 = get_point(handle, True)
        print "pt: " + str(_pt15)
    if (opts & 0x40):
        _pt16 = get_point(handle, True)
        print "pt: " + str(_pt16)
    if (opts & 0x80):
        _d40 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d40
    if (opts & 0x100):
        _d50 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d50
    if (opts & 0x200):
        _d51 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d51
    if (opts & 0x400):
        _d52 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d52
    if (opts & 0x800):
        _d53 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d53

def vport_reader(handle, flags, opts):
    print "vport_reader() ..."
    _pt10 = get_point(handle, True)
    print "pt: " + str(_pt10)
    _d40 = struct.unpack('<d', handle.read(8))[0]
    print "double: %g" % _d40
    _d41 = struct.unpack('<d', handle.read(8))[0]
    print "double: %g" % _d41
    _w68 = struct.unpack('<h', handle.read(2))[0]
    print "word: %d" % _w68
    
entities = [
    None,
    line_reader,
    point_reader,
    circle_reader,
    shape_reader,
    None,
    None,
    text_reader,
    arc_reader,
    trace_reader,
    None,
    solid_reader,
    block_reader,
    endblk_reader,
    insert_reader,
    attdef_reader,
    attrib_reader,
    sbend_reader,
    None,
    pline_reader,
    vertex_reader,
    None,
    face3d_reader,
    dim_reader,
    vport_reader
    ]

def read_p13_table(handle, lt): # no info on this ...
    print "read_p13_table() ..."
    _size, _nr, _start = lt
    print "size: %d, nr: %d, start: %d" % (_size, _nr, _start)
    handle.seek(_start, 0)
    for _i in range(_nr):
        print "reading from %#x" % handle.tell()
        _end = handle.tell() + _size
        handle.seek(_end)
    _crc32 = handle.read(32)
    
def read_dimstyle_table(handle, lt):
    print "read_dimstyle_table() ..."
    _size, _nr, _start = lt
    print "size: %d, nr: %d, start: %d" % (_size, _nr, _start)
    handle.seek(_start, 0)
    for _i in range(_nr):
        print "reading from %#x" % handle.tell()
        _end = handle.tell() + _size
        _flag = struct.unpack('B', handle.read(1))[0]
        print "flag: %#x" % _flag
        _name = handle.read(32)
        print "name: %s" % _name
        _word = struct.unpack('<h', handle.read(2))[0]
        print "word: %d" % _word
        _d4048 = struct.unpack('<9d', handle.read(72))
        print "d4048: " + str(_d4048)
        _d140145 = struct.unpack('<6d', handle.read(48))
        print "d140145: " + str(_d140145)
        _b7078 = struct.unpack('<7B', handle.read(7))
        print "b7078: " + str(_b7078)
        _b170175 = struct.unpack('<6B', handle.read(6))
        print "b170175: " + str(_b170175)
        _s3 = handle.read(16)
        print "string: %s" % _s3
        _s4 = handle.read(16)
        print "string: %s" % _s4
        _s5 = handle.read(32)
        print "string: %s" % _s5
        _s6 = handle.read(32)
        print "string: %s" % _s6
        _s7 = handle.read(32)
        print "string: %s" % _s7
        handle.seek(3, 1)
        _w176, _w177, _w178 = struct.unpack('<3h', handle.read(6))
        print "word: %d" % _w176
        print "word: %d" % _w177
        print "word: %d" % _w178
        _d146, _d147 = struct.unpack('<dd', handle.read(16))
        print "double: %d" % _d146
        print "double: %d" % _d147
        _offset = _end - handle.tell() - 2
        print "offset: %d" % _offset
        if _offset:
            handle.seek(_offset, 1)
        _crc = struct.unpack('<h', handle.read(2))[0]
        print "crc: %#x" % _crc
    _crc32 = handle.read(32)

def read_appid_table(handle, lt):
    print "read_appid_table() ..."
    _size, _nr, _start = lt
    print "size: %d, nr: %d, start: %d" % (_size, _nr, _start)
    handle.seek(_start, 0)
    for _i in range(_nr):
        print "reading from %#x" % handle.tell()
        _end = handle.tell() + _size
        _flag = struct.unpack('B', handle.read(1))[0]
        print "flag: %#x" % _flag
        _name = handle.read(32)
        print "name: %s" % _name
        _word = struct.unpack('<h', handle.read(2))[0]
        print "word: %d" % _word
        _offset = _end - handle.tell() - 2
        print "offset: %d" % _offset
        if _offset:
            handle.seek(_offset, 1)
        _crc = struct.unpack('<h', handle.read(2))[0]
        print "crc: %#x" % _crc
    _crc32 = handle.read(32)

def read_vport_table(handle, lt):
    print "read_vport_table() ..."
    _size, _nr, _start = lt
    print "size: %d, nr: %d, start: %d" % (_size, _nr, _start)
    handle.seek(_start, 0)
    for _i in range(_nr):
        print "reading from %#x" % handle.tell()
        _end = handle.tell() + _size
        _flag = struct.unpack('B', handle.read(1))[0]
        print "flag: %#x" % _flag
        _name = handle.read(32)
        print "name: %s" % _name
        _used = struct.unpack('<h', handle.read(2))[0]
        _pt10 = get_point(handle, False)
        print "pt: " + str(_pt10)
        _pt11 = get_point(handle, False)
        print "pt: " + str(_pt11)
        _pt17 = get_point(handle, True)
        print "pt: " + str(_pt17)
        _pt16 = get_point(handle, True)
        print "pt: " + str(_pt16)
        _d50, _d40 = struct.unpack('<dd', handle.read(16))
        print "double: %g" % _d50
        print "double: %g" % _d40
        _pt12 = get_point(handle, False)
        print "pt: " + str(_pt12)
        _d41, _d42, _d43, _d44 = struct.unpack('<4d', handle.read(32))
        print "double: %g" % _d41
        print "double: %g" % _d42
        print "double: %g" % _d43
        print "double: %g" % _d44
        _w7178 = struct.unpack('<8h', handle.read(16))
        print "words: " + str(_w7178)
        _d51 = struct.unpack('<d', handle.read(8))[0]
        print "double: %g" % _d51
        _pt13 = get_point(handle, False)
        print "pt: " + str(_pt13)
        _pt14 = get_point(handle, False)
        print "pt: " + str(_pt14)
        _pt15 = get_point(handle, False)
        print "pt: " + str(_pt15)
        _offset = _end - handle.tell() - 2
        print "offset: %d" % _offset
        if _offset:
            handle.seek(_offset, 1)
        _crc = struct.unpack('<h', handle.read(2))[0]
        print "crc: %#x" % _crc
    _crc32 = handle.read(32)
    
def read_ucs_table(handle, lt):
    print "read_ucs_table() ..."
    _size, _nr, _start = lt
    print "size: %d, nr: %d, start: %d" % (_size, _nr, _start)
    handle.seek(_start, 0)
    for _i in range(_nr):
        print "reading from %#x" % handle.tell()
        _end = handle.tell() + _size
        _flag = struct.unpack('B', handle.read(1))[0]
        print "flag: %#x" % _flag
        _name = handle.read(32)
        print "name: %s" % _name
        _used = struct.unpack('<h', handle.read(2))[0]
        print "used: %d" % _used
        _pt10 = get_point(handle, True)
        print "pt: " + str(_pt10)
        _pt11 = get_point(handle, True)
        print "pt: " + str(_pt11)
        _pt12 = get_point(handle, True)
        print "pt: " + str(_pt12)
        _offset = _end - handle.tell() - 2
        print "offset: %d" % _offset
        if _offset:
            handle.seek(_offset, 1)
        _crc = struct.unpack('<h', handle.read(2))[0]
        print "crc: %#x" % _crc
    _crc32 = handle.read(32)
    
def read_view_table(handle, lt):
    print "read_view_table() ..."
    _size, _nr, _start = lt
    print "size: %d, nr: %d, start: %d" % (_size, _nr, _start)
    handle.seek(_start, 0)
    for _i in range(_nr):
        print "reading from %#x" % handle.tell()
        _end = handle.tell() + _size
        _flag = struct.unpack('B', handle.read(1))[0]
        print "flag: %#x" % _flag
        _name = handle.read(32)
        print "name: %s" % _name
        _used, _db40 = struct.unpack('<hd', handle.read(10))
        print "used: %d" % _used
        print "double: %d" % _db40
        _pt10 = get_point(handle, False)
        print "pt: " + str(_pt10)
        _db41 = struct.unpack('<d', handle.read(8))[0]
        print "double: %d" % _db41
        _pt11 = get_point(handle, True)
        print "pt: " + str(_pt11)
        _pt12 = get_point(handle, True)
        print "pt: " + str(_pt12)
        _w71, _db42, _db43, _db44, _db50 = struct.unpack('<hdddd', handle.read(34))
        print "word: %d" % _w71
        print "double: %g" % _db42
        print "double: %g" % _db43
        print "double: %g" % _db44
        print "double: %g" % _db50
        #
        # things probably are messed up here
        #
        _offset = _end - handle.tell()
        print "offset: %d" % _offset - 2
        if _offset:
            handle.seek(_offset, 1)
        _crc = struct.unpack('<h', handle.read(2))[0]
        print "crc: %#x" % _crc
    _crc32 = handle.read(32)

def read_ltype_table(handle, lt):
    print "read_ltype_table() ..."
    _size, _nr, _start = lt
    print "size: %d, nr: %d, start: %d" % (_size, _nr, _start)
    handle.seek(_start, 0)
    for _i in range(_nr):
        print "reading from %#x" % handle.tell()
        _end = handle.tell() + _size
        _flag = struct.unpack('B', handle.read(1))[0]
        print "flag: %#x" % _flag
        _name = handle.read(32)
        print "name: %s" % _name
        _w1 = struct.unpack('<h', handle.read(2))[0]
        print "word: %d" % _w1
        _s1 = handle.read(48)
        print "string: %s" % _s1
        _b1, _b2 = struct.unpack('BB', handle.read(2))
        print "b1: %#x; b2: %#x" % (_b1, _b2)
        _doubles = struct.unpack('<13d', handle.read(104))
        print "doubles: " + str(_doubles)
        _crc = struct.unpack('<h', handle.read(2))[0]
        print "crc: %#x" % _crc
        _offset = _end - handle.tell()
        print "offset: %d" % _offset
        if _offset:
            handle.seek(_offset, 1)
    _crc32 = handle.read(32)
    
def read_style_table(handle, lt):
    print "read_style_table() ..."
    _size, _nr, _start = lt
    print "size: %d, nr: %d, start: %d" % (_size, _nr, _start)
    handle.seek(_start, 0)
    for _i in range(_nr):
        print "reading from %#x" % handle.tell()
        _end = handle.tell() + _size
        _flag = struct.unpack('B', handle.read(1))[0]
        print "flag: %#x" % _flag
        _name = handle.read(32)
        print "name: %s" % _name
        _w1, _d1, _d2, _d3, _b1, _d4 = struct.unpack('<hdddBd', handle.read(35))
        print "word: %d" % _w1
        print "doubles: (%g,%g,%g)" % (_d1, _d2, _d3)
        print "byte: %#x" %_b1
        print "double: %g" % _d4
        _s1 = handle.read(128)
        print "string: %s" % _s1
        _crc = struct.unpack('<h', handle.read(2))[0]
        print "crc: %#x" % _crc
        _offset = _end - handle.tell()
        print "offset: %d" % _offset
        if _offset:
            handle.seek(_offset, 1)
    _crc32 = handle.read(32)
    
def read_layer_table(handle, lt):
    print "read_layer_table() ..."
    _size, _nr, _start = lt
    print "size: %d, nr: %d, start: %d" % (_size, _nr, _start)
    handle.seek(_start, 0)
    for _i in range(_nr):
        print "reading from %#x" % handle.tell()
        _end = handle.tell() + _size
        _flag = struct.unpack('B', handle.read(1))[0]
        print "flag: %#x" % _flag
        _name = handle.read(32)
        print "name: %s" % _name
        _used, _color, _style, _crc = struct.unpack('<hhhh', handle.read(8))
        print "used: %d" % _used
        print "color: %d" % _color
        print "style: %d" % _style
        print "crc: %#x" % _crc        
        _offset = _end - handle.tell()
        print "offset: %d" % _offset
        if _offset:
            handle.seek(_offset, 1)
    _crc32 = handle.read(32)
    
def read_block_table(handle, bt):
    print "read_block_table() ..."
    _size, _nr, _start = bt
    print "size: %d, nr: %d, start: %d" % (_size, _nr, _start)
    handle.seek(_start, 0)
    for _i in range(_nr):
        print "reading from %#x" % handle.tell()
        _end = handle.tell() + _size        
        print "record: %d" % _i
        _flag = struct.unpack('B', handle.read(1))[0]
        print "flag: %#x" % _flag
        _name = handle.read(32)
        print "name: %s" % _name
        _used, _b1, _w1, _b2, _w2, _crc = struct.unpack('<hBhBhh', handle.read(10))
        print "used: %d" % _used
        print "b1: %#x" % _b1
        print "w1: %d" % _w1
        print "b2: %#x" % _b2
        print "w2: %d" % _w2
        print "crc: %#x" % _crc
        _offset = _end - handle.tell()
        print "offset: %d" % _offset
        if _offset:
            handle.seek(_offset, 1)
    _crc32 = handle.read(32)

def read_xdata(handle):
    _data = []
    _xlen = struct.unpack('<h', handle.read(2))[0]
    while (_xlen > 0):
        _xval = struct.unpack('B', handle.read(1))[0]
        _xlen = _xlen - 1
        if _xval == 0:
            if _xlen < 1: break
            _val = struct.unpack('B', handle.read(1))[0]
            _xlen = _xlen - 1
            if r13: # ugh
                if _xlen < 1: break
                _cp = struct.unpack('B', handle.read(1))[0] # code page?
                _xlen = _xlen - 1
            if _xlen < _val: break
            _string = handle.read(_val)
            _data.append(_string)
            _xlen = _xlen - _val
        elif (_xval == 1 or
              _xval == 3 or
              _xval == 70):
            if _xlen < 2: break
            _val = struct.unpack('<h', handle.read(2))[0]
            _data.append(_val)
            _xlen = _xlen - 2
        elif _xval == 2:
            if _xlen < 1: break
            _val = struct.unpack('B', handle.read(1))[0]
            if _val == 0:
                _data.append('{')
            elif _val == 1:
                _data.append('}')
            else:
                raise ValueError, "Unexpected char: %#02x" % _val
            _xlen = _xlen - 1
        elif _val == 5:
            if _xlen < 8: break
            _val = handle.read(8) # probably an entity handle
            _xlen = _xlen - 8
        elif (_xval == 40 or
              _xval == 41 or
              _xval == 42):
            if _xlen < 8: break
            _val = struct.unpack('<d', handle.read(8))[0]
            _data.append(_val)
            _xlen = _xlen - 8
        elif (_xval == 10 or
              _xval == 11 or
              _xval == 12 or
              _xval == 13): # 13 not in R12 spec sheet; is in R13/R14 spec
            if _xlen < 24: break
            _point = get_point(handle, True)
            _data.append(_point)
            _xlen = _xlen - 24
        elif _xval == 71:
            if _xlen < 4: break
            _val = struct.unpack('<l', handle.read(4))[0]
            _data.append(_val)
            _xlen = _xlen - 4
        else:
            _xlen = 0
    return _data
    
def read_entities(handle, start, end):
    print "read_entities() ..."
    handle.seek(start, 0)
    _at = handle.tell()
    _emax = len(entities)
    while (_at < (end - 32)):
        print "handle offset: %d [%#x]" % (_at, _at)
        print "max offset: %d [%#x]" % ((end - 32), (end - 32))
        _kind, _flag, _length = struct.unpack("<BBh", handle.read(4))
        print "kind: %#x" % _kind
        print "flag: %#x" % _flag
        print "length: %d" % _length
        _crcpos = _at + (_length - 2)
        print "crcpos: %d [%#x]" % (_crcpos, _crcpos)
        _layer, _opts = struct.unpack('<hh', handle.read(4))
        print "layer: %d" % _layer
        print "opts: %d" % _opts
        _color = 0
        if (_flag & 0x1):
            _color = struct.unpack('B', handle.read(1))[0]
        print "color: %#x" % _color
        _extra = 0
        if (_flag & 0x40):
            _extra = struct.unpack('B', handle.read(1))[0]
        print "extra: %#x" % _extra
        _xdata = None
        if (_extra & 0x2):
            _xdata = read_xdata(handle)
        _type = None
        if (_flag & 0x2):
            _type = struct.unpack('<h', handle.read(2))[0]
            print "type: %#x" % _type
        _z = None
        if ((_flag & 0x4) and (_kind > 2) and (_kind != 22)):
            _z = struct.unpack('<d', handle.read(8))[0]
            print "z: %g" % _z
        _th = None
        if (_flag & 0x8):
            _th = struct.unpack('<d', handle.read(8))[0]
            print "th: %g" % _th
        _handle = None
        if (_flag & 0x20):
            _handle = get_handle(handle)
            print "handle: " + str(_handle)
        _paper = None
        if (_extra & 0x4):
            _paper = struct.unpack('<h', handle.read(2))[0]
            print "paper: %d" % _paper
        if _kind < _emax:
            _reader = entities[_kind]
            if _reader is not None:
                _reader(handle, _flag, _opts)
                _at = handle.tell()
                if (_at < _crcpos):
                    print "at < crcpos"
                    handle.seek((_crcpos - _at), 1)
                elif (_at > _crcpos):
                    print "at > crcpos"
                else:
                    print "at == crcpos"
                _crc = struct.unpack('<h', handle.read(2))[0]
                print "crc: %d [%#x]" % (_crc, _crc)
        _at = handle.tell()
    _crc32 = handle.read(32)
    print "entity reading completes at %#x" % handle.tell()
        
def read_header(handle):
    print "read_header() ..."
    _w1 = struct.unpack("<h", handle.read(2))[0]    
    print "word: %d" % _w1
    _inbase = get_point(handle, True)
    print "inbase: " + str(_inbase)
    _extmin = get_point(handle, True)
    print "extmin: " + str(_extmin)
    _extmax = get_point(handle, True)
    print "extmax: " + str(_extmax)
    _limmin = get_point(handle, False)
    print "limmin: " + str(_limmin)
    _limmax = get_point(handle, False)
    print "limmax: " + str(_limmax)
    _vcx, _vcy, _d3, _d4 = struct.unpack("<dddd", handle.read(32))
    print "view center: (%g, %g)" % (_vcx, _vcy)
    print "doubles: %g, %g" % (_d3, _d4)
    _b1, _b2 = struct.unpack("<BB", handle.read(2))
    print "bytes: %#02x, %#02x" % (_b1, _b2)
    _sx, _sy = struct.unpack("<dd", handle.read(16))
    print "snap units: (%g,%g)" % (_sx, _sy)
    _b56 = array.array('B')
    _b56.fromfile(handle, 56)
    print "56 bytes: " + str(_b56)
    _d7, _d8, _d9 = struct.unpack("<ddd", handle.read(24))
    print "doubles: %g, %g, %g" % (_d7, _d8, _d9)
    _b18 = array.array('B')
    _b18.fromfile(handle, 18)
    print "18 bytes: " + str(_b18)
    _d10 = struct.unpack("<d", handle.read(8))
    print "double: %g" % _d10
    #
    # the following comes from the C code
    #
    _at = handle.tell()
    print "handle offset: %d [%#x]" % (_at, _at)
    _w2, _w3 = struct.unpack("<hh", handle.read(4))
    print "word: %d, %d" % (_w2, _w3)
    _b44 = array.array('B')
    _b44.fromfile(handle, 44)
    print "44 bytes: " + str(_b44)
    _at = handle.tell()
    print "handle offset: %d [%#x]" % (_at, _at)
    _b354 = array.array('B')
    _b354.fromfile(handle, 354)
    print "354 more bytes read ..."
    if _b354[0:4] == [0x61,0x63,0x61,0x64]: # acad
        print "found 'acad' at start of 354 byte block"
    _at = handle.tell()
    print "handle offset: %d [%#x]" % (_at, _at)
    #
    # skip to 0x3ef - remainder of header is mysterious ...
    #
    handle.seek(0x3ef, 0)

def get_point(handle, flag):
    if flag:
        _pt = struct.unpack('<ddd', handle.read(24))
    else:
        _pt = struct.unpack('<dd', handle.read(16))
    return _pt

def get_table(handle):
    return struct.unpack("<hll", handle.read(10))    

def get_table_test(handle):
    return struct.unpack('<hhhl', handle.read(10))

def get_handle(handle):
    _bytes = []
    _len = struct.unpack('B', handle.read(1))[0]
    for _i in range(_len):
        _byte = struct.unpack('B', handle.read(1))[0]
        _bytes.append(_byte)
    return (5, _len) + tuple(_bytes)

def get_string(handle):
    _len = struct.unpack('<h', handle.read(2))[0]
    return handle.read(_len)
        
if __name__ == '__main__':
    if len(sys.argv) < 2:
        print "usage: dwg12.py filename"
        sys.exit(1)
    try:
        _fh = file(sys.argv[1])
    except:
        print "failed to open '%s'. Exiting ..." % sys.argv[1]
        sys.exit(1)
    dump_info(_fh)
    _fh.close()
